#!/usr/bin/env python
# coding: utf-8

import h5py
import numpy as np
import matplotlib.pyplot as plt
#from fast_python.baseline import baseline
from scipy import ndimage
from ..utils.misc import extend_Trues

def mask_rfi_p(T2p, s_sigma=10, chan_smooth_method='gaussian', times_s=1, times=6, ext_add=0, ext_frac=0):
    """
    T2p: array, shape=(x,N,2); T or flux with two polar and sorted by time
    s_sigma: int; gaussian smooth size along time (axis=0)
    chan_smooth_method: method of smoothing along channel axis
    times_s: # for smoothed (XX-YY)
    times: # for origin (XX-YY)
    """
    if s_sigma is not None: 
        T2p_s = ndimage.gaussian_filter1d(T2p, s_sigma, axis=0)
    else:
        T2p_s = T2p

    diff = T2p_s[:,:,0] - T2p_s[:,:,1]
    # smooth channel
    chan_smooth_method = 'gaussian'
    chan_smooth_sigma = 5
    if chan_smooth_method == 'gaussian':
        diff_s = ndimage.gaussian_filter1d(diff, chan_smooth_sigma, axis=1)
    else:
        diff_s = np.zeros_like(diff)
        for i in range(len(diff)):
            diff_s[i]= baseline(x=None, y=diff[i], method = 'arPLS', lam=400, deg=3, offset=2)[1]
    # residual
    resi = diff - diff_s
    vdown = np.nanpercentile(resi, 10, axis=1)
    vmean = np.nanpercentile(resi, 50, axis=1)
    vupper = np.nanpercentile(resi, 90, axis=1)
    # find rfi
    # compare smoothed diff
    is_rfi = diff_s < (vmean - times_s*(vmean - vdown))[:,None]
    is_rfi = is_rfi | (diff_s > (vmean + times_s*(vupper - vmean))[:,None])
    # compare origin diff
    is_rfi = is_rfi | (diff < (vmean - times*(vmean - vdown))[:,None])
    is_rfi = is_rfi | (diff > (vmean + times*(vupper - vmean))[:,None])
    # relatively diff
    diff_frac = diff_s/np.nanmean(T2p_s,axis=2)
    is_rfi_and= abs(diff_frac) > 0.25
    diff_frac_2 = diff/np.nanmean(T2p_s,axis=2)
    is_rfi_and= is_rfi_and | (abs(diff_frac) > 0.25)
    
    is_rfi = is_rfi & is_rfi_and
    
    if ext_add > 0 or ext_frac > 0:
        is_rfi = extend_Trues(is_rfi, axis=1, ext_add=ext_add, ext_frac=ext_frac)
    return is_rfi



#!/usr/bin/env python
# coding: utf-8

import numpy as np
from copy import deepcopy
from hifast.utils.io import BaseIO

def vopt2vrad(vopt):
    """
    velocity: optical to radio
    """
    c = 299792.458
    vrad = c -c**2/(c + vopt)
    return vrad

def vrad2vopt(vrad):
    """
    velocity: radio to optical
    """
    c = 299792.458
    vopt = c**2/(c - vrad) - c
    return vopt

def redshift(v , relative = False):
    c = 299792.458
    beta = v / c
    
    if relative:
        g = 1 / np.sqrt(1 - beta ** 2)
        z = (1 + v / c) * g - 1
        return z
    else:
        return betas

def percent_vminmax(data,percent = None):
    if percent != None:
        vmin = np.nanpercentile(data,q = (1-percent)/2* 100,interpolation='nearest')
        vmax = np.nanpercentile(data,q = (1+percent)/2* 100,interpolation='nearest')
        return vmin,vmax
    else:
        return np.nanmin(data),np.nanmax(data)

def _round_up_to_odd_integer(value):
    c = np.ceil(value)
    if np.isscalar(c):
        if c % 2 == 0:
            r = c + 1
        else:
            r = c
    else:
        r = np.zeros(c.shape[0])
        for i in range(c.shape[0]):
            if c[i] % 2 == 0:
                r[i] = c[i] + 1
            else:
                r[i] = c[i]
    return r.astype('int')

def date2mjd(date):
    """
    date: str; UTC+8
    """
    from astropy.time import Time; import astropy.units as u
    t = Time(date, format='iso', scale='utc') - 8*u.hour
    
    return t.mjd

def do_smooth(s1p, s_method_t = 'none', s_sigma_t = None, 
              s_method_freq = 'none', s_sigma_freq = None, is_rfi = None):
    T = deepcopy(s1p)
    if is_rfi is None: is_rfi = np.full(T.shape[:2], False, dtype=bool)
    is_excluded = np.all(is_rfi,axis = 1)
    T[is_rfi] = 0

    from ..utils.misc import smooth1d
    if s_method_t in ['gaussian', 'boxcar', 'median']:
        print('Smoothing t ...')
        T[~is_excluded] = smooth1d(T[~is_excluded], axis=0, sigma=s_sigma_t, method=s_method_t)
    elif s_method_t == 'iter_median':
        print('Smoothing t ...')
        from .sw_fft import med_fit_ripple
        T[~is_excluded] = med_fit_ripple(T[~is_excluded], nspec = s_sigma_t, func='iter')

    if s_method_freq in ['gaussian', 'boxcar', 'median']:
        print('Smoothing freq ...')
        T[~is_excluded] = smooth1d(T[~is_excluded], axis=1, sigma=s_sigma_freq, method=s_method_freq)
    return T

def do_smooth_onoff(s1p, is_on = None,is_rfi = None, **kwargs):
    T = deepcopy(s1p)
    if is_on is None: 
        T = do_smooth(T,is_rfi = is_rfi, **kwargs)
    else:
        if is_rfi is None: is_rfi = np.full(T.shape[:2], False, dtype=bool)
        T[is_on] = do_smooth(T[is_on],is_rfi = is_rfi[is_on], **kwargs)
        T[~is_on] = do_smooth(T[~is_on],is_rfi = is_rfi[~is_on], **kwargs)
    return T

def line_set(ax,xlabel,ylabel,direction='in',xlim=None,ylim=None,legend=True,
             title=None,loc='best', size = None,frameon=True):
    """
    plot settings
    """
    if size is None:
        size = {'mz': 1,   # Set thickness of the tick marks
                'lz': 3,   # Set length of the tick marks
                'lbz': 14,  # Set label size
                'tkz': 12,  # Set tick size
                }
    mz = size['mz']; lz = size['lz']
    lbz = size['lbz']; tkz = size['tkz']
    # Make tick lines thicker
    for l in ax.get_xticklines():
        l.set_markersize(lz)
        l.set_markeredgewidth(mz)
    for l in ax.get_yticklines():
        l.set_markersize(lz)
        l.set_markeredgewidth(mz)

    # Make figure box thicker
    for s in ax.spines.values():
        s.set_linewidth(mz)
    ax.minorticks_on()
    ax.tick_params("both",which = 'both',direction=direction,labelsize=tkz,
                  bottom=True, top=True,left=True,right=True)
    if xlim is not None: ax.set_xlim(xlim)
    if ylim is not None: ax.set_ylim(ylim)
    ax.set_xlabel(xlabel,fontsize=lbz)
    ax.set_ylabel(ylabel,fontsize=lbz)
    if legend: ax.legend(loc = loc,fontsize=tkz, frameon=frameon)
    if title is not None: ax.set_title(title,fontsize=lbz)

    
class Args(object):
    def __init__(self, fpath, frange = None, outdir = None,):
        self.fpath = fpath
        self.outdir = outdir
        self.frange = frange

class Read_hdf5(BaseIO):
    ver = 'old'
    
    def __init__(self, args, dict_in=None, inplace_args=False, HistoryAdd=None):
        """
        args: class
              including attributes: fpath, outdir, frange
        dict_in: if None, load data from args.fpath, if set, omit data in args.fpath
        """
        self.args = args if inplace_args else deepcopy(args)
        self.dict_in = dict_in
        self.HistoryAdd = HistoryAdd
        self._gen_fpath_out()
        if self.dict_in is None:
            self._check_fout()
        self.nB = self.get_nB(self.args.fpath)
        self._import_m()
        self.open_fpath()
        self.load_specs()
        self.load_radec()
        try:
            self.load_and_add_Header()
        except TypeError:
            pass
    
    def get_data(self, polar = 'none'):
        data = deepcopy(self.s2p)
        if data.shape[0] == 2 or data.shape[0] == 1:
            from ..utils.io import PolarMjdChan_to_MjdChanPolar
            data = PolarMjdChan_to_MjdChanPolar(data)
        if len(data.shape) == 3:
            if polar == 'xx':
                data = data[:,:,0]
            elif polar == 'yy':
                data = data[:,:,1]
            elif polar == 'average':
                data = np.mean(data,axis = 2)
        return data
    
    def plot_waterfall(self,data = None,xtype = 'freq', cmap = 'rainbow',per_vmin_max = None, polar = None, 
                       vmin_max = None,xylim = None,outdir = './',xrange = None,interp_method = 'nearest',
                       plot = True,pdf = None,time_label = False, figsize=(15,4),title = None,**kwargs):
        from matplotlib import pyplot as plt
        #import matplotlib
        #matplotlib.rcParams['image.interpolation'] = 'none'

        if pdf is not None:
            plt.switch_backend('agg')

        restfreq = 1420.405751#7667
        c = 299792.458
        freq = self.freq
        if xtype == 'freq':
            x = self.freq
        elif xtype == 'vrad':
            if 'vel' in self.fs.keys():
                x = self.fs['vel'][()]
            else:
                x = c*(restfreq-freq)/restfreq
        elif xtype == 'vopt':
            x = c*(restfreq-freq)/freq
        
        if not isinstance(data,np.ndarray):
            data = self.get_data(polar)
        
        if len(data.shape) != 2:
            raise ValueError(f"Check your input data shape {data.shape}. Are they 2D? ")

        if xrange != None:
            x1,x2 = np.min(xrange),np.max(xrange)
            is_use = (x>=x1)&(x<=x2)
            x = x[is_use]
#             if data.shape[1] != x.shape[0]:
#                 data = data[:,is_use]
        else:
            xrange = [x[0],x[-1]]

        if plot:
            if time_label:
                mjds = self.mjd
                extent = (xrange[0],xrange[1],mjds[0],mjds[-1]) 
            else:
                extent = (xrange[0],xrange[1],0,data.shape[0])            

            fig,ax = plt.subplots(figsize=figsize)

            if vmin_max != None:
                im=ax.imshow(data,vmin=vmin_max[0],vmax=vmin_max[1],origin='lower', cmap = cmap,
                             aspect='auto',extent = extent,interpolation=interp_method,)
            else:
                if per_vmin_max == None:   
                    im=ax.imshow(data,origin='lower', cmap = cmap,aspect='auto',extent = extent,
                                 interpolation=interp_method,)
                else:
                    vmin,vmax = percent_vminmax(data,percent = per_vmin_max)
                    im=ax.imshow(data,vmin=vmin,vmax=vmax,origin='lower', cmap = cmap,
                                aspect='auto',extent = extent,interpolation=interp_method,)

            ax.set_xlabel(xtype)
            ax.set_ylabel("specs")
            #ax.tick_params(labelsize=14)    
            ax.set_title(title)
            plt.colorbar(im,pad=.01)
            plt.tight_layout()
            if xylim is not None:
                xlim1,xlim2,ylim1,ylim2 = xylim
                ax.set_xlim(xlim1,xlim2)
                ax.set_ylim(ylim1,ylim2)
            ax.minorticks_on()
            if time_label:
                from astropy.time import Time; import astropy.units as u
                labels_new = ((Time(ax.axes.get_yticks(), format='mjd')) + 8*u.hour).strftime('%H:%M')
                ax.axes.set_yticklabels(labels_new)

            if pdf is not None:
                pdf.savefig();plt.close()
            else:
                plt.show()






__all__ = ['IO']


import numpy as np
from .utils.io import *
from .bld import IO as bld_IO


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=formatter_class, allow_abbrev=False,
                        description='Fit and subtract baseline', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input spectra temperature or flux file path.')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')
parser.add_argument('--no_radec', action='store_true', not_in_write_out_config_file=True,
                    help="don't check or add ra dec")
parser.add_argument('--show_prog', type=bool_fun, choices=[True, False], default='True', env_var='HIFAST_SHOW_PROG',
                    help='')

# baseline fitting


group = parser.add_argument_group(f'*BaseLine (set --method as none to skip this) \n{sep_line}')
group.add_argument('--nproc', '-n', type=int, default=1,
                   help='number of process used in fitting baseline')

method = ['none']
method += ['MedMed', 'MinMed']

group.add_argument('--method', default='MedMed', choices=method,
                   help='method used to fit baseline, if set as none, skip this')


# MinMed or MedMed
group = parser.add_argument_group(f'*MinMed or MedMed\n{sep_line}')
group.add_argument('--npart', type=int,
                   help='divide data into n parts')
group.add_argument('--nsection', type=int,
                   help='divide each part into n sections. Same with nspec')
group.add_argument('--nspec', type=int,
                   help='divide each part into n sections. One section has n specs.')


group = parser.add_argument_group(f'*Use pre-determined is_excluded array in input file\n{sep_line}')
group.add_argument('--use_pre_is_excluded', type=bool_fun, choices=[True, False], default='True',
                   help='If True, use pre-determined is_excluded array if existed to mask channels')

# Exclude files
group = parser.add_argument_group(f'*Exclude known source catalogs\n{sep_line}')
group.add_argument('--continuum_file',
                   help='txt catalog with #ra[deg], dec[deg], R[arcmin], freq_min[MHz], freq_max[MHz]')
group.add_argument('--src_file',
                   help='txt catalog with #ra[deg], dec[deg], R[arcmin], freq_min[MHz], freq_max[MHz]')
group.add_argument('--frame', choices=['BARYCENT', 'HELIOCEN', 'LSRK', 'LSRD'], default='LSRK',
                   help='Velocity Rest Frames, HELIOCEN or LSRK')

# low-order polynomial on individual spectrum
group = parser.add_argument_group(f'*Post-applying low-order polynomial fitting on individual spectrum if `--s_method_t` or `--njoin` used [optional]\n{sep_line}')

rew_type = ['asym1', 'asym2', 'asym3', 'sym1',]
method = ['none']
method += ['poly-'+r for r in rew_type]

group.add_argument('--post_method', default='none', choices=method,
                   help='method used to fit baseline')
group.add_argument('--post_s_method_freq', default='none', choices=['none', 'median', 'gaussian', 'boxcar', 'PLS', 'fft'],
                   help='method used to smooth each spectrum along frequency axis')
group.add_argument('--post_s_sigma_freq', type=float, default=5,
                   help='used with --post_s_method_freq')
group.add_argument('--post_average_every_freq', type=int, default=0,
                   help='bin the channels by this factor')
group.add_argument('--post_deg', type=int, default=2,
                   help='polynomial degree for `*-poly` methods')
group.add_argument('--post_offset', type=float, default=2,
                   help='baseline fit parameters')
group.add_argument('--post_ratio', type=float, default=0.01,
                   help='baseline fit parameters')
group.add_argument('--post_niter', type=int, default=100,
                   help='baseline fit parameters')
group.add_argument('--post_exclude_add', '--post_exclude_type', default='none', choices=['none', 'auto1', 'auto2'],
                   help='baseline fit parameters')


class IO(bld_IO):

    def _get_fpart(self,):
        fpart = '-ref'
        if self.args.post_method is not None and self.args.post_method != 'none':
            fpart += '_p'
        return fpart

    def _import_m(self,):
        """
        need modify this function
        """
        super()._import_m()
        global h5py, OrderedDict, sub_baseline

        import h5py
        from collections import OrderedDict
        from .core.baseline import sub_baseline


    def _load_continuum(self):
        args = self.args
        fpath = args.continuum_file
        if fpath is not None:
            from .core.regions import mask_srcs

            is_excluded = np.zeros(self.s2p.shape[:2], dtype = bool)
            print(f"loading excluded files from {fpath}")
            self.is_continum = mask_srcs(fpath, is_excluded, self.ra, self.dec, self.freq,
                                         self.mjd, inplace=True, rest_frame=args.frame)

    def med_fit_baseline(self,):
        args = self.args
        # gen self.s2p_out
        s2p = self.s2p[:]
        from copy import deepcopy
        s2p_ori = deepcopy(s2p)

        is_excluded = deepcopy(self.is_excluded) if hasattr(self,'is_excluded') else np.zeros(self.s2p.shape[:2], dtype = bool)
        for key in ['is_rfi', 'is_continum']:
            if hasattr(self, key):
                is_excluded |= getattr(self, key)
        s2p[is_excluded] = np.nan

        from .ripple.running_median import minmed
        kwargs = {}
        keys = ['nsection', 'nspec', 'npart', 'method']
        for key in keys:
            kwargs[key] = getattr(args, key)
        s2p = s2p_ori - minmed(s2p, **kwargs)
        return s2p

    def gen_s2p_out(self,):
        args = self.args

        self._load_is_rfi()
        # is_excluded
        s2p = self.s2p[:]
        # is_excluded
        self._load_is_excluded()
        self._load_sources()

        # fit baseline:
        if 'Med' in args.method:
            self._load_continuum()

            print(f'Use {args.method} to substract baseline. Remember another linear substraction.')
            s2p = self.med_fit_baseline()
        else:
            print('no baseline method assigned, skip. Make sure the input spectra have been baselined.')

        if args.post_method is not None and args.post_method !='none':
            if hasattr(self,'is_excluded'):
                is_excluded = self.is_excluded
                # check is_excluded shape
                if is_excluded.ndim == 2:
                    is_excluded = np.full(is_excluded.shape + s2p.shape[2:], is_excluded[..., None])
                else:
                    raise(ValueError('Now `is_excluded` only supports 2-dim.'))
            else:
                is_excluded = None
            # not use is_continum on baseline fitting
            print('applying low-order polynomial fitting on each spectrum')
            s2p = self.post_fit_baseline(s2p, self.freq, self.mjd, args, is_excluded)
        self.s2p_out = s2p

    def __call__(self, save=True):
        self.gen_s2p_out()
        if hasattr(self,'is_continum'):
            print("save is_continum :D")
            self.gen_dict_out(is_continum = self.is_continum)
        else:
            self.gen_dict_out()
#             if 'is_continum' in self.dict_out.keys():
#                 args = self.args
#                 self.dict_out['is_continum'] = h5py.ExternalLink(os.path.relpath(
#                     args.fpath, os.path.dirname(self.fpath_out)), f'/S/is_continum')
        if hasattr(self,'is_excluded'):
            # update
            self.dict_out['is_excluded'] = self.is_excluded
        if hasattr(self, 'is_rfi'):
            # update
            self.dict_out['is_rfi'] = self.is_rfi
        # save to hdf5 file
        if save:
            self.save()



if __name__ == '__main__':
    args_ = parser.parse_args()

    print('#'*35+'Args'+'#'*35)
    args_from = parser.format_values()
    print(args_from)
    print('#'*35+'####'+'#'*35)

    HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
    io = IO(args_, HistoryAdd=HistoryAdd)
    io()



#!/usr/bin/env python
# coding: utf-8
"""
hifast.find
===========

Search for spectra near a target coordinate.

This module searches for beams within a specified radius of a target RA/DEC in a list of HDF5 files.
"""
import os
from glob import glob
import sys
from .utils.io import ArgumentParser, add_common_argument, argparse

class Find_point(object):
    def __init__(self, p_str, r, p_u=None, r_u=None):
        self._import_m()
        if p_u is None:
            p_u = (u.hourangle, u.deg)
        if r_u is None:
            r_u = u.deg

        if ':' in p_str:
            # Sexagesimal format (HH:MM:SS DD:MM:SS) -> Hourangle, Deg
            self.point = SkyCoord(p_str, unit=p_u)
        else:
            try:
                # Try auto-detect units (e.g., "30.3deg 40.44deg", "12h30m 45d")
                self.point = SkyCoord(p_str)
            except Exception:
                # Fallback to decimal degrees (e.g., "30.3 40.44")
                self.point = SkyCoord(p_str, unit=(u.deg, u.deg))
        
        self.r = r*u.deg

    def _import_m(self,):
        """
        Lazy import of heavy dependencies.
        """
        global h5py, np, SkyCoord, u
        import h5py
        import numpy as np
        from astropy.coordinates import SkyCoord
        from astropy import units as u

    @staticmethod
    def get_radec(fname):
        # Ensure imports are available if called statically
        if 'h5py' not in globals():
             global h5py, np, SkyCoord, u
             import h5py
             import numpy as np
             from astropy.coordinates import SkyCoord
             from astropy import units as u

        f = h5py.File(fname,'r')
        S = f["S"] if "S" in f.keys() else f
        ra_list = []
        beam_list = []
        for key in S.keys():
            if 'ra' == key[:2]:
                try:
                    beam_list += [key[2:]]
                except:
                    beam_list += ['']
                ra_list += [S[key][()]]

        dec_list = [S[f'dec{beam}'][()] for beam in beam_list]
        radec = {}
        for ra, dec, beam in zip(ra_list, dec_list, beam_list):
            radec[beam] = SkyCoord(ra=ra*u.deg, dec=dec*u.deg)
        f.close()
        return radec
    def find_in(self, fname):
        radec = self.get_radec(fname)
        for beam in radec.keys():
            # Handle both scalar and array coordinates
            sep = radec[beam].separation(self.point)
            if sep.isscalar:
                sep = np.atleast_1d(sep)
            
            inds = np.where(sep < self.r)[0]
            if len(inds)>0:
                print(f'Beam {beam:>2} in', fname)
                print(inds)

def create_parser():
    parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                            allow_abbrev=False,
                            description='Search for spectra within a radius of a target coordinate.')
    add_common_argument(parser)
    
    parser.add_argument('fnames', nargs='+', metavar='FILE',
                        help='Input HDF5 file paths (supports glob patterns like "data/*.hdf5").')
    
    parser.add_argument('-p', required=True, metavar='RA_DEC',
                        help='Target coordinate string. Supports "HH:MM:SS DD:MM:SS" (e.g., "0:48:26.39 +42:34:08"), decimal degrees (e.g., "30.3 40.44"), or explicit units (e.g., "30.3deg 40.44deg").')
    
    parser.add_argument('-r', type=float, default=1, metavar='ARCMIN',
                        help='Search radius [arcmin].')
    return parser

parser = create_parser()

if __name__ == '__main__':
    args = parser.parse_args()
    fnames = args.fnames
    p = args.p
    r = args.r
    files=[]
    for thefname in fnames:
        files+= glob(thefname, recursive=True)
    files= list(set(files))
    if len(files)==0:
        import sys
        print('cannot find file')
        sys.exit()
    files.sort()
    #print(files)
    F = Find_point(p, r/60)
    [F.find_in(fname) for fname in files]





__all__ = ['download_to_temp_and_move', 'get_file', 'download_and_extract_zip']


import os
import shutil
import tempfile
import requests
import fcntl  # For file locking
import time   # For sleep during retry

def download_to_temp_and_move(url, save_path, resume=False):
    """
    Downloads a file from a URL to a temporary file, then moves it to the final location
    upon successful completion.

    Args:
        url (str): The URL of the file to download.
        save_path (str): The full path where the file should be saved.
        resume (bool, optional): If True, try to resume an existing partial download. Defaults to False.

    Returns:
        bool: True if the download was successful (or resumed successfully), False otherwise.
    """

    temp_dir = os.path.dirname(save_path)
    with tempfile.NamedTemporaryFile(delete=False, prefix=os.path.basename(save_path)+'.', dir=temp_dir) as temp_file:
        temp_path = temp_file.name

        headers = {}
        existing_size = 0
        if resume and os.path.exists(temp_path):
            existing_size = os.path.getsize(temp_path)
            headers['Range'] = f'bytes={existing_size}-'

        try:
            response = requests.get(url, stream=True, headers=headers)

            if response.status_code == 200:
                total_size = int(response.headers.get('content-length', 0))
            elif response.status_code == 206 and resume:
                content_range = response.headers.get('content-range')
                if content_range:
                    total_size = int(content_range.split('/')[-1])
                else:
                    total_size = 0
            else:
                print(f"Download failed: status code {response.status_code}")
                return None

            os.makedirs(os.path.dirname(temp_path), exist_ok=True)

            with open(temp_path, 'ab' if resume else 'wb') as file:
                downloaded_size = existing_size
                for chunk in response.iter_content(chunk_size=1048576):
                    file.write(chunk)
                    downloaded_size += len(chunk)

                    # Progress indication (optional)
                    print(f"\rDownloaded {downloaded_size}/{total_size} bytes...", end='')

            print(f"\nFile downloaded successfully to: {temp_path}")  # Print temp path

            # Move the temporary file to the final destination
            shutil.move(temp_path, save_path)
            print(f"File moved to: {save_path}")

            return save_path

        except requests.exceptions.RequestException as e:
            print(f"Download error: {e}")
            print(f"Failed to download file from: {url}, you may need to manually download it and put it at {save_path}.")
            if not resume:
                try:
                    os.remove(temp_path)
                except OSError:
                    pass
            return None


def get_file(url, save_path, resume=False, max_retries=6, retry_delay=10, overwrite=False, failure_marker=None):
    # --- Circuit Breaker Check ---
    if failure_marker and os.path.exists(failure_marker):
        try:
            # If failure was less than 60 seconds ago, skip
            if time.time() - os.path.getmtime(failure_marker) < 60:
                # print(f"Skipping download due to recent network failure: {failure_marker}")
                return None
        except OSError:
            pass

    # check if file already exists
    if not overwrite and os.path.exists(save_path):
        print(f"File already exists at: {save_path}")
        return save_path
    # mkdir if dirname doesn't exist
    os.makedirs(os.path.dirname(save_path), exist_ok=True)
    lock_file_path = save_path + ".lock"
    itsme = True
    for attempt in range(max_retries):
        lock_file = open(lock_file_path, 'w')
        try:
            fcntl.lockf(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)

            if not overwrite and os.path.exists(save_path):
                print(f"File already exists at: {save_path}")
                return save_path
            else:
                if itsme or overwrite:
                    res = download_to_temp_and_move(url, save_path, resume)
                    
                    # --- Circuit Breaker Update ---
                    if failure_marker:
                        try:
                            if res:
                                if os.path.exists(failure_marker):
                                    os.remove(failure_marker)
                            else:
                                with open(failure_marker, 'w') as f:
                                    f.write(str(time.time()))
                        except OSError:
                            pass
                    return res
                else:
                    # Should not reach here if non-blocking lock acquired
                    print(f"Another process has tried to download '{save_path}' and failed.",
                          f"You may need to manually download it from {url} and put it at {save_path}.")
                    return None
        except BlockingIOError:  # Lock is held by another process
            itsme = False
            if attempt < max_retries - 1:
                # If we are waiting for a lock, it means someone else is downloading.
                # If we wanted to OVERWRITE, we might still want to do it after they finish? 
                # Or if we just want "latest", getting their version is fine?
                # For manifest: if someone else is updating, their update is fresh enough. We can accept it.
                # So if locking conflict occurs, we treat it as "someone else handled it".
                # BUT, `get_file` retries. When it gets the lock eventually, `os.path.exists` will be true.
                # If `overwrite=True`, we would download AGAIN.
                # This ensures we get OUR version? No, for manifest, if A updates, B waiting -> B gets lock -> B updates again.
                # This is redundant but consistent with "overwrite=True".
                if attempt == 0:
                    print(f"Another process is downloading '{save_path}'. Retrying in {retry_delay} seconds...")
                time.sleep(retry_delay)
            else:
                print(f"Another process is downloading '{save_path} and max retries reached.'. Skipping...")
                return None
        finally:
            try:
                # check if the file can be locked by me
                # only remove lock file if it's created by me
                fcntl.lockf(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
                fcntl.lockf(lock_file, fcntl.LOCK_UN) # Release the lock
                lock_file.close()
                os.remove(lock_file_path)
            except:
                pass

def download_and_extract_zip(url, target_dir, resume=False, max_retries=6, retry_delay=10, failure_marker=None):
    """
    Downloads a ZIP file and extracts it to the target directory atomically.
    Ensures that the target directory is only created if extraction is successful.
    
    Args:
        url (str): The URL of the ZIP file.
        target_dir (str): The directory where the contents should be extracted.
        resume (bool): Whether to resume download.
        max_retries (int): Number of retries for lock acquisition.
        retry_delay (int): Delay in seconds between retries.
        failure_marker (str, optional): Path to circuit breaker file.
        
    Returns:
        str or None: path to target_dir if successful, None otherwise.
    """
    if os.path.exists(target_dir):
        return target_dir

    # 1. Download ZIP to a temporary location
    # Use get_file to handle locking and downloading safely
    # We append .zip to the directory name to create a temp zip path
    # But get_file expects a target path.
    # We'll use a temp directory for the zip file
    
    zip_target_path = os.path.join(os.path.dirname(target_dir), f".tmp_{os.path.basename(target_dir)}.zip")
    
    downloaded_zip = get_file(url, zip_target_path, resume=resume, max_retries=max_retries, retry_delay=retry_delay, failure_marker=failure_marker)
    
    if not downloaded_zip:
        return None
        
    # 2. Extract to a temp dir
    temp_extract_dir = os.path.join(os.path.dirname(target_dir), f".tmp_extract_{os.path.basename(target_dir)}")
    if os.path.exists(temp_extract_dir):
         shutil.rmtree(temp_extract_dir)
         
    try:
        import zipfile
        with zipfile.ZipFile(downloaded_zip, 'r') as zip_ref:
            zip_ref.extractall(temp_extract_dir)
            
        # 3. Atomic Move
        # If target_dir appeared in the meantime (race condition), we should check again
        if os.path.exists(target_dir):
             shutil.rmtree(temp_extract_dir)
             return target_dir
             
        os.rename(temp_extract_dir, target_dir)
        
        # Cleanup zip file
        try:
            os.remove(downloaded_zip)
        except OSError:
            pass
            
        return target_dir
        
    except Exception as e:
        print(f"Error extracting zip: {e}")
        if os.path.exists(temp_extract_dir):
            shutil.rmtree(temp_extract_dir)
        return None



#!/usr/bin/env python
# coding: utf-8

from .core.tcal_onoff import tcal_onoff
from .utils.io import save_specs_hdf5
from glob import glob
import numpy as np
import os
import re


# In[13]:


# nB=1
# fname_part= f'/data/inspur_disk06/fast_data/3047/M31_45_0/20191227/M31_45_0_MultiBeamOTF-M{nB:02d}_F_'


# In[432]:


def get_mjds(fname_part, outdir=None, read_all=False, t_step=1):
    mjd_delta= t_step*np.array(1.16508454084396362304687500e-05, dtype=np.float64)
    end_file= len(glob(fname_part+'*.fits'))
    
    if read_all:
        mjds=tcal_onoff(fname_part,1,end_file).mjds
    else:
        mjds_1=tcal_onoff(fname_part,1,1).mjds
        mjds_2=tcal_onoff(fname_part,end_file,end_file).mjds
        mjds= np.arange(mjds_1[0],mjds_2[-1]+mjd_delta,mjd_delta)
    
    if outdir is None:
        return {'mjd':mjds}
    else:
        outname= os.path.join(outdir, 
                f"{os.path.basename(fname_part)[:-1]}-specs_T-mjd-{os.path.basename(os.path.dirname(fname_part))}.hdf5")
        save_specs_hdf5(outname,{'mjd':mjds})


# In[478]:


def main(basedirs,outdir='./', band=None, **kargs):
    if band is None:
        band = ['F','W','N']
    band = ','.join(band)
    fnames=[]
    for basedir in basedirs:
        fnames+= glob(basedir+f'/**/*M01_[{band}]*0001.fits',recursive=True)
    fnames=list(set(fnames))
    fnames.sort()
    [print(i) for i in fnames]
    for fname in fnames:
        fname_part= re.sub("[0-9]{4}\.fits\Z",'',fname)
        print(fname_part)
        get_mjds(fname_part,outdir, **kargs)


# In[467]:


if __name__ == '__main__':
    import os
    import argparse
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument('basedirs',nargs='+',
                        help='basedirs')
    parser.add_argument('--outdir', required=False, default='./',
                       help='output file name, full path')
    parser.add_argument('-a', '--read_all', action='store_true',
                       help='read all files to obtain mjd')
    parser.add_argument('--t_step', type=float, default=1,
                       help='time step; s')
    parser.add_argument('-b', '--band', nargs='+', choices=['F', 'W', 'N'],
                       help='freq band type, F, W, N')
    args = parser.parse_args()
    main(args.basedirs, args.outdir, args.band, read_all=args.read_all, t_step=args.t_step)



#!/usr/bin/env python
# coding: utf-8
# Author: Nekomata  zmliu@nao.cas.cn
#修改nBs获取方式

from ast import arg
import numpy as np

import h5py
import os
from glob import glob
import copy
import re
from astropy.time import Time
from astropy.coordinates import SkyCoord
from astropy import units as u
import scipy.interpolate as interp
from scipy.optimize import curve_fit

#for hifast@dev
from hifast.core.cal import CalOnOff, FastRawSpec
from hifast.core.radec import get_radec
from hifast.utils.io import MjdChanPolar_to_PolarMjdChan
from hifast.utils.io import PolarMjdChan_to_MjdChanPolar
from hifast.utils.misc import smooth1d
from hifast.utils.io import save_dict_hdf5
from hifast.utils.io import replace_nB
from hifast.utils.io import ArgumentParser, bool_fun, formatter_class
from hifast.utils.io import sub_patten

import sys


global nB
global radec
global fit_k
global rlim_src # max allowed radius[arcsec] to src
global freq
global obsdate
global outname
global flux
global nproc
global outdir
nproc = 1
 

def load_crd(obj,crds):
    if crds==None:
        from astroquery.simbad import Simbad
        try:
            sbo = Simbad.query_object(obj)
            crd = SkyCoord((sbo['RA'][0]+sbo['DEC'][0]),unit=(u.hourangle,u.deg))
        except:
            print('Calibrator can not find in Simbad.query_object, please input coordinates using ``--crd``.')
            sys.exit(1)
    else:
        crd= SkyCoord(*crds,unit=(u.deg))
    return crd

import csv

def load_flux_profile(calname,freq_key,fpparas):
    """
    freq_key should in MHz
    calname should be string
    
    """
    freq_key= freq_key/1000
    if fpparas is not None:
        a0,a1,a2,a3= fpparas
    else:
        # import pandas as pd
        # fpf= pd.read_csv(os.path.dirname(__file__) + '/data/FluxProfiles.csv')
        try:
            fpath = os.path.dirname(__file__) + '/data/FluxProfiles.csv'
            found = False
            with open(fpath, 'r') as f:
                reader = csv.DictReader(f)
                for row in reader:
                    if row['name'] == calname:
                        a0 = float(row['a0'])
                        a1 = float(row['a1'])
                        a2 = float(row['a2'])
                        a3 = float(row['a3'])
                        found = True
                        break
            if not found:
                 raise ValueError("Calibrator name not found in FluxProfiles.csv")

            # narg= np.where(fpf['name']==calname)[0][0]
            # a0,a1,a2,a3= fpf['a0'][narg],fpf['a1'][narg],fpf['a2'][narg],fpf['a3'][narg]
        except:
            print('Can not find calibrator name in FluxProfile.csv. Please check calname or input flux profile parameters.')
            sys.exit(1)
    logS = a0+a1*np.log10(freq_key)+a2*(np.log10(freq_key))**2+a3*(np.log10(freq_key))**3
    flux= 10**(logS)
    return flux

#fitting method of Scan modes
def fit_curve(T,ra,figname=None):
    rcon= (ra>-0.025)&(ra<0.025)
    rcoff= ~((ra>-0.1)&(ra<0.1))
    tmax= np.max(T)
    tsys= np.mean(T[rcoff])
    if figname is not None:
        from matplotlib import pyplot as plt
        plt.figure(figsize=(8,4))
        plt.plot(ra, T, color='k')
        plt.scatter(ra[rcon], T[rcon], marker='.',color='r')
        plt.scatter(ra[rcoff],T[rcoff], marker='.',color='b')
        plt.xlabel('RA [deg]')
        plt.ylabel('T')
        plt.grid()
        plt.tight_layout()
        plt.savefig(figname)
    
    onbound=([tmax-0.05*np.abs(tmax)-tsys,-0.01,0,tsys-0.05*np.abs(tsys),],[tmax+0.05*np.abs(tmax)-tsys,0.01,1,tsys+0.05*np.abs(tsys),])
    onp,oncon = curve_fit(fit_scon,ra[rcon],T[rcon],bounds=onbound)
    offbound= ([-1,tsys-0.1*np.abs(tsys)],[1,tsys+0.1*np.abs(tsys)])
    offp,offcon = curve_fit(fit_scoff,ra[rcoff],T[rcoff],bounds=offbound)
    #ton= onp[0]+onp[3]
    #toff= offp[1]
    #return np.array([ton,toff,tmax,onp[1]])
    return tmax, onp, offp 
def fit_scon(x,a,b,c,d):
    return a*np.exp(-(x-b)**2/(2*c**2))+d
def fit_scoff(x,k,b):
    return k*x+b

#separate On Off time of Track modes
def load_track_time(mjd,t_src,nB,obsmode,n_cir):
    if obsmode=='OnOff':
        t_change = 30/60/60/24
        t_cir= t_src+t_change
        mjds= mjd[0]+np.arange(n_cir)*t_cir
        is_src,is_ref = np.full(len(mjd), False),np.full(len(mjd), False)
        for st in mjds:
            is_src = is_src | ((mjd > st) & (mjd < st+t_src))
            is_ref = is_ref | ((mjd > st+t_src+t_change) & (mjd < st+2*t_src+t_change))
        if nB==1:
            mjd_src,mjd_ref= is_src,is_ref
        else:
            mjd_src,mjd_ref= is_ref,is_src
    elif obsmode=='MultiBeamCalibration':
        t_change = 40/60/60/24
        t_cir= t_src+t_change
        mjds= mjd[0]+np.arange(20)*t_cir
        obslist= np.array([1,2,3,4,5,6,7,19,8,9,10,11,12,13,14,15,16,17,18])
        mjdon= mjds[np.where(obslist==nB)[0][0]+1]
        mjd_src= (mjdon<mjd)&(mjdon+t_src>mjd)                  
        if nB==1:
            mjd_src= mjd_src|(mjd<(mjds[0]+t_src))
        mjd_ref= (mjdon-t_cir>mjd)|(mjdon+t_cir<mjd)
    return mjd_src,mjd_ref

#mask rfi roughly
def rfi_mask(T,plot=False):
    Tsmt= smooth1d(T,'gaussian_fft',sigma_nchan,axis=0)
    Tdiff= np.abs(np.diff(Tsmt,axis=0))
    is_rfi= np.full(T.shape,False)
    dlimit= 0.015
    k= 500
    T_tmp= []
    for i in range(2):
        Tdiff_use=Tdiff[:,i]
        is_rfi_use= is_rfi[:,i]
        nlist= np.where(Tdiff_use>dlimit)[0]
        for j in range(len(nlist)):
            begin= nlist[j]-k
            end= nlist[j]+k
            if nlist[j]+1 in nlist:
                end= nlist[j]+k+1
            is_rfi_use[begin:end]= True
        Ttmp= T[~is_rfi_use,i]
        Ttmpitp= interp.interp1d(freq[~is_rfi_use],Ttmp, kind='linear', axis=0,fill_value ='extrapolate')(freq)
        T_tmp.append(Ttmpitp[:,np.newaxis])
        is_rfi[:,i]= is_rfi_use
    T_itp= np.concatenate(T_tmp,axis=1)
    T_itpsmt= smooth1d(T_itp,'gaussian_fft',sigma_nchan*10,axis=0)
    return T_itpsmt
def rfi_mask_pcal(p_cal,plot=False):
    pdiff= np.abs(np.diff(p_cal,axis=0))
    is_rfi= np.full(p_cal.shape,False)
    pdiffm= np.median(pdiff,axis=0)
    k= 500
    p_tmp= []
    for i in range(2):
        pdiff_use=pdiff[:,i]
        is_rfi_use= is_rfi[:,i]
        nlist= np.where(pdiff_use>pdiffm[i]*10)[0]
        for j in range(len(nlist)):
            begin= nlist[j]-k
            end= nlist[j]+k
            if nlist[j]+1 in nlist:
                end= nlist[j]+k+1
            is_rfi_use[begin:end]= True
        ptmp= p_cal[~is_rfi_use,i]
        ptmpitp= interp.interp1d(freq[~is_rfi_use],ptmp, kind='linear', axis=0,fill_value ='extrapolate')(freq)
        p_tmp.append(ptmpitp[:,np.newaxis])
        is_rfi[:,i]= is_rfi_use
    p_itp= np.concatenate(p_tmp,axis=1)
    return p_itp

# separate spectrum of calibrator
def sep_spe(fname,d,m,n):
    fname_part= fname
    para = {}
    para['n_delay'] = d
    para['n_on'] = m
    para['n_off'] = n
    para['start'] = 1
    para['stop'] = None
    para['frange'] = (frange[0], frange[1])
    para['verbose'] = True
    para['smooth'] ='gaussian'
    para['s_para'] = {'s_sigma':sigma}
    para['dfactor'] = None
    para['med_filter_size'] = None
    para['noise_mode'] = noise_mode
    para['noise_date'] = noise_date
    spec = CalOnOff(fname_part, **para)
    return spec

def load_pcal(pcal_s,mjd):
    mjd= mjd[:(pcal_s.shape[0])]
    if obsmode in ['Drift','DriftWithAngle','DecDriftWithAngle','MultiBeamOTF']:
        pcal_smt= np.mean(pcal_s,axis=0)
    else:
        ptime= load_track_time(mjd,t_src,nB,obsmode,n_cir=ncir)
        pcal_smt= np.mean(pcal_s[ptime[1]],axis=0)
    pcal_smt= smooth1d(pcal_smt,'gaussian_fft',sigma_nchan,axis=0)
    return pcal_smt

def plot_radec_near_cbr(ra, dec, ra_cbr, dec_cbr, figname, r_max=0.15):

    from matplotlib import pyplot as plt
    plt.figure(figsize=(5,5))
    
    is_use = (ra < ra_cbr+r_max ) & (ra > ra_cbr-r_max)
    is_use &= (dec < dec_cbr+r_max ) & (dec > dec_cbr-r_max)
    
    plt.scatter(ra[is_use], dec[is_use], marker='.', s=1, color='k')
    plt.scatter(ra_cbr, dec_cbr, marker='.', s=7, color='r')
    plt.xlabel('RA [deg]')
    plt.ylabel('DEC [deg]')
    plt.grid()
    plt.tight_layout()
    plt.savefig(figname)

def load_Tsc(spec,T,mjd):
    global radec
    global fit_k
    global rlim_src # max allowed radius[arcsec] to src
    if nB==1:
        print(f'Calculating RA-DEC for Beam {list(nBs)}')
        radec = get_radec(mjd, guess_str=fname,tol=5,nBs=list(nBs),nproc=nproc)
        save_dict_hdf5(outname+'-radec.hdf5',radec) 
        print('Saved RA-DEC to '+outname+'-radec.hdf5')
    else:
        try:
            _,_ = radec[f'ra{nB}'], radec[f'dec{nB}']
        except:
            print(f'Calculating RA-DEC for Beam {nB:02d}')
            radec = get_radec(mjd, guess_str=fname,tol=5,nBs=[nB,])
    ra0,dec0 = radec[f'ra{nB}'], radec[f'dec{nB}']
    
    radec_figname = outname + f'-M{nB:02d}-radec_near_src.png'
    plot_radec_near_cbr(ra0, dec0, ccrd.ra.deg, ccrd.dec.deg, figname=radec_figname, r_max=0.15)
    
    if obsmode in ['Drift','DriftWithAngle','DecDriftWithAngle','MultiBeamOTF']:
        T= smooth1d(T,'gaussian_fft',sigma_nchan,axis=1)
        ra= ra0-ccrd.ra.deg
        dec= dec0-ccrd.dec.deg
        data_cut= (ra>-0.5)&(ra<0.5)&(dec<0.05)&(dec>-0.05)
        Tuse= T[data_cut]
        rause= ra[data_cut]
        ONp= []
        OFFp= []
        TMAX= []
        freq_key= np.arange(frange[0]+1,frange[1],1)
        fit_k= 0
        for i in range(len(freq_key)):
            f_cut= (freq>freq_key[i]-1)&(freq<freq_key[i]+1)
            T_fit= np.mean(Tuse[:,f_cut,:],axis=1)
            try:
                if i==0:
                    figname = outname + f'-M{nB:02d}-freq_{freq_key[i]:.2f}.pdf'
                else:
                    figname = None
                tmaxXX, onpXX, offpXX  = fit_curve(T_fit[:,0],rause,figname=figname)
                tmaxYY, onpYY, offpYY  = fit_curve(T_fit[:,1],rause,figname=figname)
            except:                
                print('Fit failed in '+str(int(freq_key[i])) + f'MHz in beam {nB}, please check.')
                tmaxXX, onpXX, offpXX= np.nan,[np.nan]*4,[np.nan]*2
                tmaxYY, onpYY, offpYY= np.nan,[np.nan]*4,[np.nan]*2
                fit_k+=1
                if fit_k>=int(0.1*len(freq_key)):
                    print(f'Fit failed at too many Freq in beam {nB}. Please check')
                    sys.exit(1)
            ONp.append(np.array([onpXX,onpYY],dtype='float64'))
            OFFp.append(np.array([offpXX,offpYY],dtype='float64'))
            TMAX.append(np.array([tmaxXX,tmaxYY],dtype='float64'))
        ONp= np.array(ONp,dtype='float64')
        OFFp= np.array(OFFp,dtype='float64')
        TMAX= np.array(TMAX,dtype='float64')
        Tsrc= interp.interp1d(freq_key, ONp[:,:,0]+ONp[:,:,3], kind='linear', axis=0,fill_value ='extrapolate')(freq)
        Tref= interp.interp1d(freq_key, OFFp[:,:,1], kind='linear', axis=0,fill_value ='extrapolate')(freq)
        fit_k= fit_k/len(freq_key)
    else:
        crds= SkyCoord(ra0,dec0,unit=u.deg)          
        mjd_src,mjd_ref= load_track_time(mjd,t_src,nB,obsmode,n_cir=ncir)
        T_src,T_ref= T[mjd_src],T[mjd_ref]
        T_select = True
        if T_select:
            print(f'Further restrict the on-source spectra within {args.src_drange}')
            Tsrc,_= radec_select(T_src,crds[mjd_src], args.src_drange)
            fit_k= 1- Tsrc.shape[0]/T_src.shape[0]
            print(f'Further restrict the off-source spectra within {args.ref_drange}')
            Tref,_= radec_select(T_ref,crds[mjd_ref], args.ref_drange)
        else:
            Tsrc,Tref= T_src,T_ref
            fit_k= 1
        Tsrc= np.mean(Tsrc,axis=0)
        Tref= np.mean(Tref,axis=0)
        ONp,OFFp,TMAX= Tsrc,Tref,np.nan
    return Tsrc,Tref,ONp,OFFp,TMAX

def radec_select(T,crds,seprange):
    seps= ccrd.separation(crds).to_value(u.arcsec)
    is_use= (seps>=seprange[0])&(seps<=seprange[1])
    num_used= T[is_use].shape[0]
    if num_used==0:
        print(f'Beam {nB:02d}: No spectra inside {seprange}, please check input parameters.')
        sys.exit(1)
    else:
        print(f'Beam {nB:02d}: Input {T.shape[0]} spectra, {num_used} spectra is in {seprange}')
    return T[is_use],is_use

def load_data(fname):
    global nB
    dataname = replace_nB(fname, nB)
    spec= sep_spe(dataname,d,m,n)
    global freq
    global sigma_nchan
    freq= spec.freq_use
    sigma_nchan =  int(sigma/(np.max(freq) - np.min(freq))*(len(freq)-1))
    mjd= spec.get_mjds()
    global obsdate
    obsdate= (Time(np.median(mjd),format='mjd').to_value('iso','date_hm'))
    obsdate= re.sub('\D','',obsdate)
    
    global outdir
    outdir = sub_patten(outdir, date=obsdate+'UTC', project=calname)
    outdir = os.path.expanduser(outdir)
    if outdir!='' and (not os.path.exists(outdir)):
        print(f'outdir {outdir} not exists. Create it now')
        os.makedirs(outdir, exist_ok=True)
    global outname
    outname= f"{outdir}/{calname}-{obsmode}-{obsdate}UTC"
    
    mjd_on = mjd[spec.inds_on]
    mjd_off = mjd[spec.inds_off]
    
    Tcal_s= spec.get_Tcal_s()
    spec.gen_out_name_base(outdir)
    spec.plot = True
    spec.sep_on_off_inds()
    p_on = spec.get_field(spec.inds_on, 'DATA',)
    p_off= spec.get_field(spec.inds_off, 'DATA', close_file=False)
    #c_on, c_off, pcal_s,_= spec.get_count_tcal(spec.inds_on,spec.inds_off)
    # used power of cal not smoothed
    pcal_s = spec._get_cal_power(spec.inds_ton, spec.inds_toff_bef, spec.inds_toff_aft)
    p_cal= load_pcal(pcal_s,mjd_on[::m])
    p_cal= rfi_mask_pcal(p_cal)
    T_off= p_off*Tcal_s/p_cal

#Save T files for checking
    if saveT==True:
        odata= {}
        odata['T']= T_off
        odata['freq']= freq
        odata['mjd']= mjd_off
        odata['pcal']= p_cal
        odata['Tcal']= Tcal_s
        odata['Tcal_file']= spec.tcal_file
        save_dict_hdf5(outname+f'-M{nB:02d}_T.hdf5',odata) 
        print('Saved the temperature data to '+outname+f'-M{nB:02d}_T.hdf5')

    Tsrc,Tref,ONp,OFFp,TMAX= load_Tsc(spec,T_off,mjd_off,)
    Tsc= Tsrc-Tref
    Tsc_rfi= rfi_mask(Tsc)
    global flux
    flux= load_flux_profile(calname,freq,fpparas)
    flux= flux[:,np.newaxis]
    K_Jy= Tsc_rfi/flux
    K_Jy= K_Jy[np.newaxis,:]
    return K_Jy,Tcal_s,spec.tcal_file,mjd_off,ONp,OFFp,TMAX



sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, 
                        allow_abbrev=False,
                        description='Processing the calibrator data', )
parser.add_argument('fname',
                    help="File path of raw data, to be used with the `--nBs` parameter: '*M01*-0001.fits'.")
parser.add_argument('--nBs', type=int, nargs='*',
                    help='beam numbers. Input int number, seperated by space, such as `1 3 11`. \
                    If no numbers are provided, process all beams when the `--obsmode` is `MultiBeamCalibration` or `MultiBeamOTF`, otherwise only beam 1 will be processed.', default=None)
parser.add_argument('--outdir', type=str,
                   help='output directory',default=None)

group = parser.add_argument_group(f'*Calibrator setting\n{sep_line}')
group.add_argument('--obsmode',type=str, choices=['MultiBeamCalibration', 'OnOff', 'MultiBeamOTF','DriftWithAngle','DecDriftWithAngle','Drift'],
                    help='Observation mode. Supports `MultiBeamCalibration`, `OnOff`, and some special scanning.',)
group.add_argument('--calname', type=str, required=True,
                   help='calibrator name. e.g. 3C48',)
group.add_argument('--crd', type=float, nargs=2,
                   help='Calibrator coordinates in deg. Two float numbers, for RA and DEC. \
                   If None, use Simbad.query_object to query by `--calname`.', default=None)

calnames = "3C48,3C286,..."
try:
    calnames_list = []
    with open(os.path.dirname(__file__) + '/data/FluxProfiles.csv', 'r') as f:
         reader = csv.DictReader(f)
         for row in reader:
             calnames_list.append(row['name'])
    if calnames_list:
        calnames = ','.join(calnames_list)
except:
    pass

# fpf= pd.read_csv(os.path.dirname(__file__) + '/data/FluxProfiles.csv')
# calnames = ','.join(fpf['name'])
group.add_argument('--fluxProfilePara', type=float, nargs=4,
                   help=f'The function of calibrator flux with respect to frequency. Function of {calnames} are already embeded in code.\
                   Others must be specified and include four floating-point numbers: a0, a1, a2, and a3. \
                   The function will be represented by the equation `a0 + a1*log10(f) + a2*log10(f)**2 + a3*log10(f)**3`.',default= None)


group = parser.add_argument_group(f'*Parameters same with hifast.sep\n{sep_line}')
group.add_argument('-d','--d', '--n_delay', type=int,
                   help='time of delay divided by sampling time')
group.add_argument('-m','--m', '--n_on', type=int,
                   help='time of Tcal_on divided by sampling time')
group.add_argument('-n','--n', '--n_off', type=int,
                    help='time of Tcal_off divided by sampling time')
group.add_argument('--frange', type=float, nargs=2,
                   help='freq range',default= [1000,1500] )
group.add_argument('--noise_mode', default='high', choices=['high','low'],
                    help='noise_mode, high or low')
group.add_argument('--noise_date', default='auto',type=str,
                    help='noise obs date, default auto')
group.add_argument('--smt_sigma', type=float,
                   help='smooth sigma (in MHz) of gaussian smooth along frequency',default=1)

group = parser.add_argument_group(f'*Parameters in MultiBeamCalibration or OnOff\n{sep_line}')
group.add_argument('--t_src', type=int,
                   help='Tracking time of On-Source in second. (It is used in conjunction with Off-Source and change time to determine the On/Off source spectra. \
                   In the `OnOff` mode, the Off-Source time used is identical to the On-Source time, with a change time of 30 seconds. \
                   In the `MultiBeamCalibration` mode, the change time is 40 seconds.)', default=60)
group.add_argument('--n_cir', '--n_repeat', type=int, default=1,
                   help='The number of on-source off-source cycles in `OnOff`. It is 1 in `MultiBeamCalibration`')
# group.add_argument('--T_select', type=bool_fun, choices=[True, False], default='True',
#                    help='Select spectra within `--rlim_src` from calibrater source if `--obsmode MultiBeamCalibration`')
group.add_argument('--src_drange', nargs=2, type=float, default=[0., 20.],
                   help='Further limit the distance of On-Source spectra to the Calibrator coordinates in this range. unit is arcsecond',)
group.add_argument('--ref_drange', nargs=2, type=float, default=[180., 3600.],
                   help='Further limit the distance of Off-Source spectra to the Calibrator coordinates in this range. unit is arcsecond',)
group = parser.add_argument_group(f'*Others\n{sep_line}')
group.add_argument('--saveT', type=bool_fun, choices=[True, False], default='False',
                   help='Save T files or not',)

if __name__ == '__main__':

    args = parser.parse_args()
    fname= args.fname
    obsfile= fname.split('/')[-1]
    print('Flux Calibration of '+obsfile)
    d, m, n = args.d, args.m, args.n
    frange= args.frange
    noise_mode = args.noise_mode
    noise_date = args.noise_date
    sigma= args.smt_sigma
    obsmode= args.obsmode
    modes= ['Drift','MultiBeamCalibration','OnOff','MultiBeamOTF','DriftWithAngle','DecDriftWithAngle']
    if obsmode not in modes:
        print('Please check observation mode.')
        sys.exit(1) 
    CalinBs= args.nBs 
    if CalinBs== None:
        if obsmode in ['MultiBeamCalibration','MultiBeamOTF',]:
            nBs= np.arange(1,20)
        else:
            nBs= [1,]
    else:
        nBs= CalinBs
    print('Calibrating beams:')
    print(nBs)
    outdir= args.outdir
    if outdir==None:
        outdir= '.'
    calname= args.calname
    ccrd = load_crd(calname,args.crd)
    fpparas= args.fluxProfilePara
    t_src= args.t_src/3600/24#change into days
    ncir= args.n_cir
    T_select=args.T_select
    rlim_src = args.rlim_src
    saveT= args.saveT
    oname= f'{outdir}/{calname}-{obsmode}-'

    outdata={}
    for i in nBs:
        global nB
        nB= i
        K_Jy,Tcal_s,tcal_file,mjd,ONp,OFFp,TMAX= load_data(fname)
        ra,dec,ZD = radec[f'ra{i}'], radec[f'dec{i}'],radec[f'ZD{i}']
        outdata[f'M{nB:02d}']= K_Jy
        outdata[f'Tcal{nB}']= Tcal_s
        outdata[f'ra{i}']= ra
        outdata[f'dec{i}']= dec
        outdata[f'ZD{i}']= ZD
        outdata[f'ONp{i}']= ONp
        outdata[f'OFFp{i}']= OFFp
        outdata[f'Tmax{i}']= TMAX
        print(f'Finished calibration of M{nB:02d}, {fit_k*100:.3f}% input data has bean ignored.')
        print('#'*50)
        
    outdata['Tcal_file']= tcal_file
    print('Used Tcal of '+tcal_file)
    outdata['freq']= freq
    outdata['cal_flux']=flux
    outdata['mjd']= mjd
    if len(nBs)==19 :
        outname= outname+'-FluxGain-All.hdf5'
    else:
        outname= outname+'-FluxGain.hdf5'
    save_dict_hdf5(outname,outdata) 
    print('Saved to '+outname)





__all__ = ['IS_Interactive', 'h5py_write', 'write_header', 'binned_statistic_e', 'split_continues', 'CheckPCal',
           'CalOnOffSav', 'CalOnOffM', 'CalOnOff1111', 'FASTRawCut']


import numpy as np
from scipy import interpolate as interp
from scipy import ndimage
import h5py

import os
import re
from glob import glob

from .cal import *
from ..utils.io import MjdChanPolar_to_PolarMjdChan, gen_carta_group, save_specs_hdf5

from astropy.stats import sigma_clip

import warnings

import __main__
IS_Interactive = not hasattr(__main__, '__file__')
if not IS_Interactive:
    import matplotlib
    matplotlib.use('Agg')
from matplotlib import pyplot as plt


import h5py
import os
def h5py_write(fname, mode='w'):
    try:
        f = h5py.File(fname, mode)
    except OSError:
        if mode == 'w':
            from datetime import datetime
            os.rename(fname, fname+'.del.empty.'+datetime.now().strftime("%Y%m%d-%H%M%S"))
            f = h5py.File(fname, mode)
    return f


def write_header(f, header):
    f.create_group('Header', track_order=False)
    for key2 in header.keys():
        f['Header'].attrs[key2] = header[key2]


def mean_a(arr):
    """
    calculate the mean value of 5th to 95th peercent of the input arr
    """
    b, e = np.percentile(arr, [5,95],)
    return np.mean(arr[(arr <=e ) & (arr >= b)], dtype='float64')

# def mean_a(arr):
#     """
#     calculate the mean value of 5th to 95th peercent of the input arr
#     """
#     return np.median(arr)


from scipy.stats import binned_statistic

def binned_statistic_e(x, values, axis=-1, **kwargs):
    """
    values: 3-Dim, stat at axis=1, stored in C-order
    """
    from collections import namedtuple
    res_tuple = namedtuple('BinnedStatisticResult',
                                   ('statistic', 'bin_edges', 'binnumber'))
    # move axis to last
    values = np.moveaxis(values, axis, -1)
    res = binned_statistic(x, values.reshape((-1, values.shape[-1])), **kwargs)
    res_stat = res[0].reshape(values.shape[:-1] + (-1,))
    # move back
    res_stat = np.moveaxis(res_stat, -1, axis)
    res = res_tuple(res_stat, *res[1:])
    return res


def test_binned_statistic_e():
    x = np.random.rand(1000)
    values = np.random.random((6, 1000, 2))
    assert (binned_statistic_e(x, values, axis=1, statistic='mean')[0][...,0] ==
            binned_statistic(x, values[...,0], statistic='mean')[0]).all()
    assert (binned_statistic_e(x, values, axis=1, statistic='mean')[0][...,1] ==
            binned_statistic(x, values[...,1], statistic='mean')[0]).all()

def test_binned_statistic_e_2():
    x = np.random.rand(1000)
    values = np.random.random((2, 6, 1000, 2))
    assert (binned_statistic_e(x, values, axis=2, statistic='mean')[0][0, ...,0] ==
            binned_statistic(x, values[0, ...,0], statistic='mean')[0]).all()
    assert (binned_statistic_e(x, values, axis=2, statistic='mean')[0][0, ...,1] ==
            binned_statistic(x, values[0, ...,1], statistic='mean')[0]).all()
    assert (binned_statistic_e(x, values, axis=2, statistic='mean')[0][1, ...,0] ==
            binned_statistic(x, values[1, ...,0], statistic='mean')[0]).all()
    assert (binned_statistic_e(x, values, axis=2, statistic='mean')[0][1, ...,1] ==
            binned_statistic(x, values[1, ...,1], statistic='mean')[0]).all()


def split_continues(bools):
    """
    bools: shape (m,); dtype bool
    """
    if bools.size == 0:
        raise('empty arr')
    diff = np.diff(bools.astype('int16'))
    ind_inv = np.where(diff != 0)[0] + 1

    ind_b = np.insert(ind_inv, 0, 0)
    ind_e = np.append(ind_inv, len(bools))-1

    return zip(bools[ind_b], ind_b, ind_e)


def test_split_continues():
    for arr_bools in [np.array([False] + [True]*3 + [False]*5 + [True]),
              np.array([True] + [False]*5 + [True] + [False]*3),
              np.array([False]*10),
             ]:
        segs = []
        for val, b, e in split_continues(arr_bools):
            seg = arr_bools[b:e+1]
            assert (seg == val).all()
            segs += [seg,]
            print(val, seg)
        assert (np.hstack(segs) == arr_bools).all()


class CheckPCal(CalOnOff):
    """
    """
    def __init__(self, *args, freq_step_c=2,
                              pcal_vary_lim_bin=0.01,
                              pcal_bad_lim_t=0.5,
                              pcal_bad_lim_freq=0.7,
                              **kwargs):
        # self.pcal_vary_frac = pcal_vary_frac
        self.pcal_vary_lim_bin = pcal_vary_lim_bin
        self.pcal_bad_lim_t = pcal_bad_lim_t # not used now
        self.pcal_bad_lim_freq = pcal_bad_lim_freq
        self.freq_step_c = freq_step_c
        super().__init__(*args, **kwargs)

    def _get_vary_frac(self, inds, freq_bins_c, bin_stat='mean'):
        """
        gen change fraction
        """
        if inds.ndim != 2:
            raise(ValueError('inds should be 2D.'))
        if bin_stat == 'mean_a':
            bin_stat = mean_a
        p = self.get_field(inds)
        ps, _, _ = binned_statistic_e(self.freq_use, p, axis=2, bins=freq_bins_c, statistic=bin_stat)
        ps_max, ps_min = np.max(ps, axis=1), np.min(ps, axis=1)
        df = (ps_max - ps_min)/ps_min
        return df


    def check_pcal(self, inds_ton, freq_step_c=2):
        """
        check pcal in freq bins
        """
        n_c = 6
        assert n_c - n_c//2 <= self.n_off

        n_cbef = n_c//2
        n_caft = n_c - n_cbef
        inds_cbef = inds_ton[:,:1] - np.arange(1,n_cbef+1)[::-1]
        inds_caft = inds_ton[:,-1:] + np.arange(1,n_caft+1)

        # fix first
        is_exceed = inds_cbef[0] < 0
        n_exceed = len(np.where(is_exceed)[0])
        if n_exceed >0:
            inds_cbef[0][is_exceed] = inds_caft[0][-1] + np.arange(1,n_exceed+1)
            inds_cbef[0][np.isin(inds_cbef[0], self.inds_on)] = inds_caft[0][-1]
            inds_cbef[0].sort()
        # fix last
        is_exceed = inds_caft[-1] > self.inds[-1]
        n_exceed = len(np.where(is_exceed)[0])
        if n_exceed >0:
            inds_caft[-1][is_exceed] = inds_cbef[-1][0] - np.arange(1,n_exceed+1)[::-1]
            inds_caft[-1][np.isin(inds_caft[-1], self.inds_on)] = inds_cbef[-1][0]
            inds_caft[-1].sort()
        inds_coff = np.hstack([inds_cbef, inds_caft])
        # been sorted !
        inds_coff[0].sort()
        inds_coff[-1].sort()

        # check freq bins
        freq_range = self.freq_use[-1] - self.freq_use[0]
        freq_delta = self.freq_use[1] - self.freq_use[0]
        bins_num = 1 + int(np.round(freq_range/freq_step_c))
        if bins_num < 2:
            bins_num = 2
        freq_bins_c = np.linspace(self.freq_use[0]-freq_delta/10, self.freq_use[-1]+freq_delta/10, bins_num)
        # check off
        is_bad_fbins = self._get_vary_frac(inds_coff, freq_bins_c) > self.pcal_vary_lim_bin
        # check on
        if inds_ton.shape[1] >= 2:
            is_bad_fbins = is_bad_fbins | (self._get_vary_frac(inds_ton, freq_bins_c) > self.pcal_vary_lim_bin)
        return freq_bins_c, is_bad_fbins

    def gen_pcals(self,):
        freq_bins_c, is_bad_fbins = self.check_pcal(self.inds_ton, self.freq_step_c)
        is_aband_whole = np.sum(is_bad_fbins, axis=1) / is_bad_fbins.shape[1] > self.pcal_bad_lim_freq
        ind_in_bins = np.digitize(self.freq_use, freq_bins_c) - 1

        pcals = self._get_cal_power(self.inds_ton, self.inds_toff_bef, self.inds_toff_aft)
        if getattr(self, 'plot_pcals', False):
            from ..waterfall import plot_im
            fig, axs = plt.subplots(2, 2, figsize=(16,10))
            plot_im(pcals[..., 0], x=self.freq_use, ax=axs[0][0], interpolation='none')
            plot_im(pcals[..., 1], x=self.freq_use, ax=axs[0][1], interpolation='none')
            axs[0][0].set_title('pcals ori, polar 0')
            axs[0][1].set_title('pcals ori, polar 1')
        fw = getattr(self, 'fw_pcals', None)
        if fw is not None:
            fw['pcals_ori'] = pcals.astype('float32')
        # replace as nan
        pcals[is_bad_fbins[:, ind_in_bins]] = np.nan
        for i in range(is_aband_whole.shape[1]):
            pcals[is_aband_whole[:,i], :, i] = np.nan
        if getattr(self, 'plot_pcals', False):
            plot_im(pcals[..., 0], x=self.freq_use, ax=axs[1][0], interpolation='none')
            plot_im(pcals[..., 1], x=self.freq_use, ax=axs[1][1], interpolation='none')
            axs[1][0].set_title('pcals masked, polar 0')
            axs[1][1].set_title('pcals masked, polar 1')
            fig.suptitle(os.path.basename(self.out_name_base + '-pcals.png'))
            fig.tight_layout()
            fig.savefig(self.out_name_base + '-pcals.png', dpi=111)
        if fw is not None:
            fw['pcals'] = pcals.astype('float32')
        self.pcals = pcals
        self.freq_bins_c = freq_bins_c
        self.is_bad_fbins = is_bad_fbins
        self.ind_in_bins = ind_in_bins
        self.is_aband_whole = is_aband_whole


class CalOnOffSav():
    def __call__(self, outdir='./', step=None, header=None, sep_save=False, save_pcals=False, cali=True):
        """
        get T, mjd etc, and save in hdf5 file

        Parameters
        ----------
        outdir : str
            output directory
        step : int
            number of chunk files to process each time
        header : dict
            add in the output file
        sep_save : bool
            if True, save file every step.
        """
        #
        self.gen_out_name_base(outdir)
        n_step = self.lens[0]*step if step is not None else len(self.inds)
        inds_range = np.append(np.arange(0, self.inds[-1], n_step), self.inds[-1]+1)
        # inds_splited= np.array_split(self.inds, len(self.inds)//n_step)
        mjds = []
        extra = self.get_extra()
        # tcal
        tc_inter = self.get_Tcal_s() if cali else 'none'
        # update and write header
        header = {} if header is None else header.copy()
        if hasattr(self, 'tcal_file'):
            header.update({'tcal_file': self.tcal_file})
        # prepare for self.get_count_tcal

        self.plot = True
        self.plot_pcals = True

        if cali:
            if save_pcals:
                outname = self.out_name_base + '-pcals.hdf5'
                fwp = h5py_write(outname)
                write_header(fwp, header)
                fwp_S = fwp.create_group('S')
                self.fw_pcals = fwp_S
            self.prepare_pcals()
#             if len(inds_ton_not_use)/(len(inds_ton_use) + len(inds_ton_not_use)) > 1/3.:
#                 warnings.warn('More than one-third of the Cals are abandoned. Please check your data and ``--pcal_vary_frac``.')
            if save_pcals:
                print(f"saving pcals in {outname}")
                fwp_S['inds_ton'] = self.inds_ton
                fwp_S['is_aband_whole'] = self.is_aband_whole
                for attr in ['pcals_merged',
                             'pcals_merged_s',
                             'pcals_amp_diff_interp_values',
                            ]:
                    if hasattr(self, attr):
                        fwp_S[attr] = getattr(self, attr)
                fwp.close()
                try:
                    del self.pcals
                except AttributeError:
                    pass

        for i in range(len(inds_range)-1):
            print('part', i)
            b, e = inds_range[i:i+2]
            inds_on = self.inds_on[(self.inds_on >= b) & (self.inds_on < e)]
            inds_off = self.inds_off[((self.inds_off >= b) & (self.inds_off < e))]
            inds = np.hstack([inds_on, inds_off])
            sort = np.argsort(inds)
            mjd = self.get_field(inds[sort], field='UTOBS')  # have sorted

            if cali:
                count_tcal_res = self.get_count_tcal(inds_on, inds_off)
            else:
                count_tcal_res = [self.get_field(inds_on, 'DATA',), self.get_field(inds_off, 'DATA', close_file=True)]

            T = np.vstack(count_tcal_res[:2])
            del count_tcal_res
            if cali:
                T = T*tc_inter
            T = T[sort]  # sort T
            T = T.astype('float32')  # finally convert to float32

            if sep_save:
                res = {}
                for key in extra.keys():
                    res[key] = extra[key][inds[sort]]
                res['mjd'] = mjd
                res['freq'] = self.freq_use
                res['Ta'] = MjdChanPolar_to_PolarMjdChan(T)
                res['Tcal'] = tc_inter
                res['inds_ton'] = self.inds_ton
                res['is_aband_whole'] = self.is_aband_whole
                for attr in ['pcals_merged',
                             'pcals_merged_s',
                            ]:
                    if hasattr(self, attr):
                        res[attr] = getattr(self, attr)
                if hasattr(self, 'pcals_amp_diff_interp_values'):
                    res['pcals_amp_diff_interp_values'] = self.pcals_amp_diff_interp_values[inds[sort]]

                # print(res)
                outname = self.out_name_base + f"-specs_T_{i:04d}_{i+1:04d}.hdf5"
                res['Header'] = header
                save_specs_hdf5(outname, res, wcs_data_name='Ta')
                print(f"Saved to {outname}")
                del res
            else:
                T = MjdChanPolar_to_PolarMjdChan(T)
                if i == 0:
                    outname = self.out_name_base + f"-specs_T.hdf5"
                    fout = h5py_write(outname)
                    #
                    fout.create_group('S')
                    g = fout['S']
                    # prepare writing spec
                    d_shape = list(T.shape)
                    d_shape[1] = len(self.inds)
                    g.create_dataset('Ta', shape=d_shape, dtype=T.dtype, chunks=True)
                g['Ta'][:, b:e, :] = T
                fout.flush()
                mjds += [mjd, ]
            self.plot = False  # only plot for the first loop
        # save other small data
        if not sep_save:
            write_header(fout, header)
            for key in extra.keys():
                g[key] = extra[key]
            g['mjd'] = np.hstack(mjds)
            g['freq'] = self.freq_use
            g['Tcal'] = tc_inter
            g['inds_ton'] = self.inds_ton
            g['is_aband_whole'] = self.is_aband_whole
            for attr in ['pcals_merged',
                         'pcals_merged_s',
                         'pcals_amp_diff_interp_values',
                        ]:
                if hasattr(self, attr):
                    g[attr] = getattr(self, attr)

            print(f"Saved to {outname}")
            gen_carta_group(fout, g['Ta'].shape, wcs_data_name='Ta', axis1=g['freq'], axis2=g['mjd'])
            fout.close()


class CalOnOffM(CheckPCal, CalOnOffSav, CalOnOff):

    def gen_squeeze_freq_is_use(self, ):
        """selecting channel used to calculate relative amp"""
        frange = self.squeeze_diff_freq_frange
        self.squeeze_freq_is_use = []
        for i in range(self.is_bad_fbins.shape[-1]):
            # first exclude the whole abandoned
            is_bad_fbins = self.is_bad_fbins[..., i][~self.is_aband_whole[:,i]]
            # select fbins with bad frac less than ...
            frac = np.sum(is_bad_fbins, axis=0)/is_bad_fbins.shape[0]
            bin_is = frac <= self.squeeze_diff_freq_bad_lim
            #bin_is = ~np.logical_or.reduce(self.is_bad_fbins[..., i][~self.is_aband_whole[:,i]], axis=0)
            is_ = bin_is[self.ind_in_bins]
            if frange is not None:
                is_ &= (self.freq_use > frange[0]) & (self.freq_use < frange[1])
            self.squeeze_freq_is_use += [is_,]

    def squeeze_freq(self, arr):
        """
        squeeze freq axis
        """
        method = self.squeeze_diff_freq

        if arr.ndim !=3 :
            raise(ValueError('need 3d'))
        if method == 'median':
            squ_fun = np.nanmedian
        elif method == 'mean':
            squ_fun = np.nanmean
        else:
            raise(ValueError(f'method {method} not supported'))
        arr_s = np.zeros((arr.shape[0], arr.shape[2]), dtype='float64')
        for i in range(arr.shape[2]):
            is_ = self.squeeze_freq_is_use[i]
            arr_s[:,i] = squ_fun(arr[:, is_, i], axis=1)
        return arr_s

    def merge_pcals(self,):
        """
        merging pcals at all time and smoothing it
        """
        method_merge = self.method_merge
        # 'scale' or 'none'
        pre_process = self.merge_cal_pre_process

        pcals = self.pcals
        if pre_process == 'scale':
            # use the power not is_bad_fbins to sacle
            amp = self.squeeze_freq(pcals)
            # not change self.pcals
            if self.calc_diff_method == 'div':
                pcals = pcals * (np.nanmedian(amp, axis=0) / amp)[:,None,:]
            elif self.calc_diff_method == 'sub':
                pcals = pcals + (np.nanmedian(amp, axis=0) - amp)[:,None,:]
            # keep scaled
            self.pcals_scaled = pcals
            self.pcals_scaled_amp = amp

        if method_merge == 'median':
            pcals_merged = np.nanmedian(pcals, axis=0, keepdims=True)
        elif method_merge == 'mean':
            pcals_merged = np.nanmean(pcals, axis=0, dtype='float64', keepdims=True)

        self.pcals_merged = pcals_merged
        self.pcals_merged_s = self._get_smoothed(pcals_merged, check_nan=True)

    def gen_pcals_amp_diff(self,):
        """
        run self.merge_pcals first to get self.pcals_merged and self.pcals_merged_s
        gen self.pcals_amp_diff

        Parameters
        ----------
        squeeze_diff_freq: str; 'median', 'mean', 'sigclip_median', 'sigclip_mean'
        exclued_franges: str; freq ranges which are not used in calculating amplitude offset; e.g. [[1375, 1385],]
        """
        pcals = self.pcals
        calc_diff_method = self.calc_diff_method

        if calc_diff_method == 'div':
            pcals_diff = pcals / self.pcals_merged_s
        elif calc_diff_method == 'sub':
            pcals_diff = pcals - self.pcals_merged_s
        else:
            raise(ValueError(f"calc_diff_method:{calc_diff_method}"))
        self.pcals_amp_diff = self.squeeze_freq(pcals_diff)
        self.pcals_amp_diff_inds = np.mean(self.inds_ton, axis=1)
        self.calc_diff_method = calc_diff_method

    def gen_pcals_amp_diff_interp_values(self,):
        """
        interplate pcals_amp_diff
        gen self.pcals_amp_diff_interp_values: the amplitude offset relate to mergred pcals, shape is (inds, polar)

        Parameters
        ----------
        method_interp: str; 'gaussian' 'CubicSpline', 'CubicHermiteSpline', kind in scipy.interpolate.interp1d
                            if 'polyXd'('X'is a integer), use poly fitting with deg=X
        edges: 'nearest', 'extrapolate'
        """
        pcals_amp_diff = self.pcals_amp_diff
        method_interp = self.method_interp
        sigma_t = self.method_interp_sigma_t
        edges = self.method_interp_edges

        self.pcals_amp_diff_interp_values = np.zeros((len(self.inds), pcals_amp_diff.shape[1]), dtype='float64')
        for i in range(pcals_amp_diff.shape[1]):
            x = self.pcals_amp_diff_inds
            y = pcals_amp_diff[:, i]
            is_ = np.isfinite(y)
            x = x[is_]
            y = y[is_]
            if method_interp == 'gaussian':
                t_sample = self.hduls[0][1].data['EXPOSURE'][0]
                sigma = sigma_t/((self.n_on+self.n_off)*t_sample)/(t_sample*self.n_on)
                y = ndimage.gaussian_filter1d(y, sigma, axis=0)
                fun = interp.interp1d(x, y, fill_value="extrapolate", kind='quadratic', axis=0)
            elif method_interp == 'CubicSpline':
                fun = interp.CubicSpline(x, y, extrapolate=None)
            elif method_interp == 'CubicHermiteSpline':
                fun = interp.CubicHermiteSpline(x, y, extrapolate=None)
            elif method_interp.startswith('poly'):
                fun = np.poly1d(np.polyfit(x, y, deg=int(method_interp[4:-1])))
            else:
                # method_interp is one of "kind"
                fun = interp.interp1d(x, y, fill_value="extrapolate", kind=method_interp)
            self.pcals_amp_diff_interp_values[:, i] = fun(self.inds)
            if edges == "nearest":
                if (self.inds < x [0]).any():
                    self.pcals_amp_diff_interp_values[self.inds < x [0], i] = fun(x[0])
                if (self.inds > x [-1]).any():
                    self.pcals_amp_diff_interp_values[self.inds > x [-1], i] = fun(x[-1])



    def set_para_pcals(self, calc_diff_method='div',
                       squeeze_diff_freq='median',
                       squeeze_diff_freq_frange=None,
                       squeeze_diff_freq_bad_lim=1,
                       method_interp='quadratic',
                       method_interp_edges='nearest',
                       method_interp_sigma_t=600,
                       method_merge='median',
                       merge_cal_pre_process='scale',
                       ):
        """
        para used in self.prepare_pcals

        calc_diff_method: str; 'div', 'sub'
        """
        self.calc_diff_method = calc_diff_method # in self.pcals_amp_interp & self.get_count_tcal
        self.squeeze_diff_freq = squeeze_diff_freq
        self.squeeze_diff_freq_frange = squeeze_diff_freq_frange
        self.squeeze_diff_freq_bad_lim = squeeze_diff_freq_bad_lim
        self.method_interp = method_interp
        self.method_interp_edges = method_interp_edges
        self.method_interp_sigma_t = method_interp_sigma_t
        self.method_merge = method_merge
        self.merge_cal_pre_process = merge_cal_pre_process

    def prepare_pcals(self,):
        self.gen_pcals()
        self.gen_squeeze_freq_is_use()
        self.merge_pcals()
        self.gen_pcals_amp_diff()
        self.gen_pcals_amp_diff_interp_values()
        self.plot_pcal_merged()
        if not IS_Interactive:
            if hasattr(self, 'pcals_scaled'):
                del self.pcals_scaled

    def plot_pcal_merged(self,):
        if not getattr(self, 'plot_pcals', False):
            return
        fig, axs = plt.subplots(3,2,figsize=(15,15))

        if hasattr(self, 'pcals_scaled'):
            pcals = self.pcals_scaled
        else:
            pcals = self.pcals
        try:
            bins = self.freq_bins_c
        except:
            try:
                bwidth = 1.2 * self.s_para['s_sigma']
            except:
                bwidth = 3.5
            bins = np.arange(self.freq_use[0], self.freq_use[-1] + bwidth, bwidth)
        pcals_d, _, _ = binned_statistic_e(self.freq_use, pcals, bins=bins, axis=1, statistic='mean')
        freq_d, _, _ = binned_statistic_e(self.freq_use, self.freq_use, bins=bins, axis=0, statistic='mean')
        ax = axs[0][0]
        ax.plot(freq_d, pcals_d[:,:,0].T, alpha=0.5)
        ax.set_title(f'pcals[_scaled] downsample, polar 0')
        ax = axs[0][1]
        ax.plot(freq_d, pcals_d[:,:,1].T, alpha=0.5)
        ax.set_title(f'pcals[_scaled] downsample, polar 1')

        ax = axs[1][0]
        ax.plot(self.freq_use, self.pcals_merged[0,:,0], 'r', alpha=0.5)
        ax.plot(self.freq_use, self.pcals_merged_s[0,:,0], 'k', lw=2)
        ax.set_title(f'pcals merged, polar 0')
        ax.set_xlabel(f'Freq')
        if hasattr(self, 'squeeze_freq_is_use'):
            ax = ax.twinx()
            ax.plot(self.freq_use, self.squeeze_freq_is_use[0])
            ax.set_ylim(-0.3, 14)
        ax = axs[1][1]
        ax.plot(self.freq_use, self.pcals_merged[0,:,1], 'r', alpha=0.5)
        ax.plot(self.freq_use, self.pcals_merged_s[0,:,1], 'k', lw=2)
        ax.set_title(f'pcals merged, polar 1')
        ax.set_xlabel(f'Freq')
        if hasattr(self, 'squeeze_freq_is_use'):
            ax = ax.twinx()
            ax.plot(self.freq_use, self.squeeze_freq_is_use[1])
            ax.set_ylim(-0.3, 14)

        ax = axs[2][0]
        ax.plot(self.inds, self.pcals_amp_diff_interp_values[:,0])
        ax.scatter(self.pcals_amp_diff_inds, self.pcals_amp_diff[:,0], color='r')
        ax.set_title('pcal amp diff, polar 0')
        ax.set_xlabel(f'Index')
        ax = axs[2][1]
        ax.plot(self.inds, self.pcals_amp_diff_interp_values[:,1])
        ax.scatter(self.pcals_amp_diff_inds, self.pcals_amp_diff[:,1], color='r')
        ax.set_title('pcal amp diff, polar 1')
        ax.set_xlabel(f'Index')
        fig.suptitle(os.path.basename(self.out_name_base))
        fig.tight_layout()
        fig.savefig(self.out_name_base + '-pcals-merged.png')


    def get_count_tcal(self, inds_on, inds_off):
        """
        run self.prepare_pcals first
        get count of tcal
        """
        #
        # interp cal power amp
        amp_interp_on = self.pcals_amp_diff_interp_values[inds_on]
        amp_interp_off = self.pcals_amp_diff_interp_values[inds_off]
        # load on and off power
        c_on = self.get_field(inds_on, 'DATA',).astype('float64')
        c_off = self.get_field(inds_off, 'DATA', close_file=True).astype('float64')
        #print(p_on.dtype, p_off.dtype, p_cal_s.dtype)
        if getattr(self, 'plot', False):
            figname = self.out_name_base + "-sep.pdf"
            plot_sep(inds_on, inds_off, c_on, c_off, figname=figname)
        # count as pcal
        if self.calc_diff_method == 'div':
            c_on /= (self.pcals_merged_s * amp_interp_on[:, None, :])
            c_on -= 1  # have subtracted cal
            c_off /= (self.pcals_merged_s * amp_interp_off[:, None, :])
        elif self.calc_diff_method == 'sub':
            c_on /= (self.pcals_merged_s + amp_interp_on[:, None, :])
            c_on -= 1  # have subtracted cal
            c_off /= (self.pcals_merged_s + amp_interp_off[:, None, :])

        return c_on, c_off


class CalOnOff1111(CheckPCal, CalOnOffSav, CalOnOff):
    @staticmethod
    def split_along_a1d2(bools_2d):
        """
        """
        if bools_2d.ndim !=2 :
            raise('should input 2d array')

        ind_0, ind_1 = np.where(np.diff(bools_2d.astype('int16'), axis=1) !=0)
        ind_inv = np.unique(ind_1) + 1
        ind_1_b = np.insert(ind_inv, 0, 0)
        ind_1_e = np.append(ind_inv, bools_2d.shape[1])-1

        return ind_1_b, ind_1_e


    def plot_pcal_s(self,):
        if not getattr(self, 'plot_pcals', False):
            return
        from ..waterfall import plot_im
        fig, axs = plt.subplots(1, 2, figsize=(16,6))
        plot_im(self.pcals_s[..., 0], x=self.freq_use, ax=axs[0], interpolation='none')
        plot_im(self.pcals_s[..., 1], x=self.freq_use, ax=axs[1], interpolation='none')
        axs[0].set_title('pcals smoothed, polar 0')
        axs[1].set_title('pcals smoothed, polar 1')
        fig.suptitle(os.path.basename(self.out_name_base))
        fig.tight_layout()
        fig.savefig(self.out_name_base + '-pcals-smoothed.png')

    def set_para_pcals(self, cal_dis_lim=2):
        self.delat_t_lim = cal_dis_lim

    def prepare_pcals(self, ):
        self.gen_pcals()

        pcals_s = self._get_smoothed(self.pcals, check_nan=True)
        del self.pcals

        segs_be_list = []
        is_use_in_segs_list = []
        for polar in range(pcals_s.shape[-1]):
            bools_2d = np.isnan(pcals_s[:, :, polar])
            segs_be = self.split_along_a1d2(bools_2d)
            if len(segs_be[0]) > 2*len(self.freq_bins_c):
                raise(ValueError('something is wrong'))
            segs_be_list.append(segs_be)
            is_use_in_segs = []
            for b, e in zip(*segs_be):
                seg = bools_2d[:, b:e+1]
                nan_num = np.sum(seg, axis=1)
                if len(np.unique(nan_num)) > 2:
                    raise(ValueError('nan values splitting failed'))
                is_use_in_segs.append(nan_num == 0)
            is_use_in_segs_list.append(is_use_in_segs)
        self.pcals_s = pcals_s
        self.segs_be_list = segs_be_list
        self.is_use_in_segs_list = is_use_in_segs_list
        self.inds_ton_m = np.mean(self.inds_ton, axis=1)
        self.plot_pcal_s()
        fw = getattr(self, 'fw_pcals', None)
        if fw is not None:
            fw['pcals_s'] = self.pcals_s.astype('float32')


    def get_count_tcal(self, inds_on, inds_off):
        delat_t_lim = self.delat_t_lim
        pcals_s = self.pcals_s

        c_on = self.get_field(inds_on, 'DATA',).astype('float64')
        c_off = self.get_field(inds_off, 'DATA', close_file=True).astype('float64')

        if getattr(self, 'plot', False):
            figname = self.out_name_base + "-sep.pdf"
            plot_sep(inds_on, inds_off, c_on, c_off, figname=figname)

        inds_ton_m = self.inds_ton_m

        for polar in range(pcals_s.shape[-1]):
            is_use_in_segs = self.is_use_in_segs_list[polar]
            segs_be = self.segs_be_list[polar]
            for i, (b, e) in enumerate(zip(*segs_be)):
                inds_ton_m_use = inds_ton_m[is_use_in_segs[i]]
                # on
                inds_close = np.argmin(abs(inds_ton_m_use - inds_on[:, None]), axis=1)
                c_on[:, b:e+1, polar] /= pcals_s[:, b:e+1, polar][is_use_in_segs[i]][inds_close]
                ## mask as nan if cal and spec not close enough
                is_too_far = abs(inds_on - inds_ton_m_use[inds_close]) > (self.n_on + self.n_off)*delat_t_lim
                c_on[is_too_far, b:e+1, polar] = np.nan

                # on
                inds_close = np.argmin(abs(inds_ton_m_use - inds_off[:, None]), axis=1)
                c_off[:, b:e+1, polar] /= pcals_s[:, b:e+1, polar][is_use_in_segs[i]][inds_close]
                ## mask as nan if cal and spec not close enough
                is_too_far = abs(inds_off - inds_ton_m_use[inds_close]) > (self.n_on + self.n_off)*delat_t_lim
                c_off[is_too_far, b:e+1, polar] = np.nan
        c_on -= 1  # have subtracted cal
        return c_on, c_off


class FASTRawCut(FastRawData):
    def __init__(self,
                 fname_part,
                 start=1,
                 stop=None,
                 frange=None,
                 verbose=False,):
        """
        fname_part: str; one of chunk file name
        start: int; chunk file start
        stop: int; chunk file stop
        frange: frequency range
        """
        # generate fits filename list
        fname_part = re.sub('[0-9]{4}\.fits\Z', '', fname_part)
        if stop is None:
            stop = len(glob(fname_part+'*.fits'))
        self.fname_part = fname_part
        filenames = [fname_part+'%04d.fits'%i for i in range(start,stop+1)]
        if len(filenames) == 0:
            raise(OSError(f"can not find file, please check fname_part:{fname_part}"))
        super().__init__(filenames, frange=frange, verbose=verbose)

        self.nB = int(re.findall(r'-M[0-1][0-9]', fname_part)[-1][2:])

        self.field_names = self.hduls[0][1].data.dtype.names

        self.fits_header = dict(self.hd1s[0].items())

    def _get_freq(self,):
        # make sure not correct to process 'FREQ_new'
        super()._get_freq(center_corr=None)

    def gen_out_name_base(self, outdir):
        self.out_name_base = os.path.join(outdir, f"{os.path.basename(self.fname_part)}")

    def gen_new_freq_axis(self, ):
        is_ = (self.freq >= self.frange[0]) & (self.freq <= self.frange[1])
        self.NCHAN_new = len(np.where(is_)[0])
        self.FREQ_new = self.freq[is_][0]
        self.is_use_chan = is_

    def write_fits_header(self, g, **kwargs):
        """
        g: h5py group
        kwargs: header field pair: name:value
        """
        for key in kwargs.keys():
            g.attrs[key] = kwargs[key]


    def load_all_fields(self, inds, chan_fix=True, polar_trans=True):
        dict1 = {} # small
        dict2 = {} # large
        for key in self.field_names:
            if key != 'DATA':
                dict1[key] = self.get_field(inds, key)
                if dict1[key].dtype.char == 'U':
                    dict1[key] = dict1[key].astype('S')
            else:
                dict2[key] = self.get_field(inds, key, close_file=True)

        if chan_fix:
            dict1['NCHAN'][:] = self.NCHAN_new
            dict1['FREQ'][:] = self.FREQ_new
            assert dict2['DATA'].shape[1] == self.NCHAN_new
        if polar_trans:
            dict2['DATA'] = MjdChanPolar_to_PolarMjdChan(dict2['DATA'])

        return dict1, dict2

    def dict_stack(self, *args):
        """
        stack item in dicts, keys must be same
        """
        res = {}
        if len(args) == 0:
            return ()
        for key in args[0].keys():
            res[key] = np.hstack([d[key] for d in args])
        return res

    def add_S_group(self, f):
        """
        add a hdf5 group for spcetra type. add 'freq', hdf5 softlink to 'DATA', 'UTOBS' in '/1'
        f: file handle
        """
        g = f.create_group('S')
        # for W,N,F. also see FastRawSpec._get_freq
        center_corr = 0.000476837158203125/2.
        g['freq'] = self.freq_use + center_corr
        g['DATA'] = h5py.SoftLink('/1/DATA')
        g['Ta'] = h5py.SoftLink('/1/DATA')
        g['mjd'] = h5py.SoftLink('/1/UTOBS')


    def __call__(self, outdir='./', step=1, header=None, sep_save=False, h5_compression='none'):
        """

        Parameters
        ----------
        outdir : str
            output directory
        step : int
            number of chunk files to process each time
        header : dict
            add in the output file
        sep_save : bool
            if True, save file every step.
        """
        if h5_compression == 'none':
            h5_compression = None
        self.gen_out_name_base(outdir)
        self.gen_new_freq_axis()

        n_step = self.lens[0]*step if step is not None else len(self.inds)
        inds_range = np.append(np.arange(0, self.inds[-1], n_step), self.inds[-1]+1)
        # inds_splited= np.array_split(self.inds, len(self.inds)//n_step)
        mjds = []
        # update and write header
        header = {} if header is None else header.copy()

        dict1_list = []
        stype = 'DATA'
        for i in range(len(inds_range)-1):
            print('part', i)
            b, e = inds_range[i:i+2]
            inds = self.inds[((self.inds >= b) & (self.inds < e))]
            dict1, dict2 = self.load_all_fields(inds, chan_fix=True, polar_trans=True)
            # large data
            T = dict2[stype]
            if sep_save:
                outname = self.out_name_base + f"{i+1:04d}.hdf5"
                fout_sep = h5py_write(outname)
                write_header(fout_sep, header)
                g_sep_0 = fout_sep.create_group('0') # table 0
                self.write_fits_header(g_sep_0, **dict(self.hd0s[i*step].items()))
                g_sep = fout_sep.create_group('1') # table 1
                for key in dict1.keys():
                    g_sep[key] = dict1[key]
                g_sep.create_dataset(stype, data=T, chunks=True, compression=h5_compression)
                # NAXIS1=self.NCHAN_new*self.2 ?
                self.write_fits_header(g_sep, **self.fits_header)
                # update new
                self.write_fits_header(g_sep,
                                       NAXIS2=len(inds),
                                       TDIM21=f"(2, {self.NCHAN_new})",
                                      )
                # waterfall
                mjd = dict1['UTOBS']
                gen_carta_group(fout_sep, g_sep[stype].shape, wcs_data_name=stype, axis1=self.freq_use, axis2=mjd, wcs_data_group='1')
                # add S group
                self.add_S_group(fout_sep)
                fout_sep.close()
                print(f"Saved to {outname}")
            else:
                # open file
                if i == 0:
                    outname = self.out_name_base + f"0001.hdf5"
                    fout = h5py_write(outname)
                    g_0 = fout.create_group('0') # table 0
                    self.write_fits_header(g_0, **dict(self.hd0s[0].items()))
                    fout.create_group('1')
                    g = fout['1']
                    # prepare writing spec
                    d_shape = list(T.shape)
                    d_shape[1] = len(self.inds)
                    g.create_dataset(stype, shape=d_shape, dtype=T.dtype, chunks=True, compression=h5_compression)
                g[stype][:, b:e, :] = T
                fout.flush()
                dict1_list.append(dict1)
        # save other small data
        if not sep_save:
            write_header(fout, header)
            dict1 = self.dict_stack(*dict1_list)
            for key in dict1.keys():
                g[key] = dict1[key]
            self.write_fits_header(g, **self.fits_header)
                # update new
            self.write_fits_header(g,
                                   NAXIS2=len(self.inds),
                                   TDIM21=f"(2, {self.NCHAN_new})",
                                   )
            # waterfall
            mjd = dict1['UTOBS']
            gen_carta_group(fout, g[stype].shape, wcs_data_name=stype, axis1=self.freq_use, axis2=mjd, wcs_data_group='1')
            # add S group
            self.add_S_group(fout)
            fout.close()
            print(f"Saved to {outname}")



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/ripple/markRFI.ipynb (unless otherwise specified).

__all__ = ['find_t', 'mask_sf', 'mask_lf', 'mask_time_rfi', 'mask_nr', 'mask_freq_rfi', 'mask_thr_rfi']

# Cell
# author: Xu Chen 2021.06

import os

from astropy import log
from matplotlib import pyplot as plt

import numpy as np
import h5py
from copy import deepcopy
from tqdm import tqdm

from .markRFI import rms, real_rms, std, real_std, get_startend, find_edge_2sides

def load_hdf5_spec(name, frange = None):
    import h5py
    
    f = h5py.File(name,'r')
    fs = f['S']
    freq = fs['freq'][()]
    if frange is None:
        frange = [0, np.inf]
    use = (freq > frange[0]) & (freq < frange[1])
    if 'T' in fs.keys():
        infield = 'T'
    elif 'Ta' in fs.keys():
        infield = 'Ta'
    elif 'flux' in fs.keys():
        infield = 'flux'
    else:
        raise(ValueError('can not find spec'))
    s2p = fs[infield][()]
    if s2p.shape[0] == 2 or s2p.shape[0] == 1:
        from ..utils.io import PolarMjdChan_to_MjdChanPolar
        s2p = PolarMjdChan_to_MjdChanPolar(s2p)
    f.close()
    return s2p[:,use,:]
    
####################### time rfi ########################

def find_t(data,freq,frange,thr_type = 'input_med_times',
           mean_times = 10,diff_times = 20,rfi_width_lim = 20,ext_add = 0,plot = False,
           pdf = None,xylim = None,plot_norfi=False,axis = 'time'):
    """
    data: 2D arrays
    frange: like [a,b]
    mean_times: mean_thr = median value * mean_times
    diff_times: diff_thr = median value * diff_times
    rfi_width_lim: width limit
    ext_add: extend edge
    """
    if (freq is not None)&(frange is not None):
        if (len(frange) == 2):
            pattern_use = (freq>frange[0])&(freq<frange[1])
            pat_data = data[:,pattern_use]
        else:
            raise ValueError("frange should like [fmin,fmax].")
    else:
        pat_data = data

#     print(f"finding rfi along {axis} axis.")
    if axis == 'time':
        pat_mean = np.nanmean(pat_data,axis = 1)
        t = np.arange(data.shape[0])
        ylabel = 'mean along freq axis'
        xlabel = 'spec number (time axis)'
    elif axis == 'freq':
        pat_mean = np.nanmean(pat_data,axis = 0)
        t = deepcopy(freq)#np.arange(data.shape[1])
        ylabel = 'mean along time axis'
        xlabel = 'freq (MHz)'

    is_timerfi = np.zeros_like(t,dtype = 'bool')

    pat_med = np.nanmedian(pat_mean)

    if pat_med <= 0 and thr_type == 'input_med_times':
        log.warning(f"Median value <= 0. Switch to thr_type = 'input_absmed_times'")
        thr_type = 'input_absmed_times'

    if thr_type == 'input_med_times':
        pass
    else:
        if thr_type == 'input_absmed_times':   
            pat_mean = np.abs(pat_mean)
        elif thr_type == 'input_posimed_times':   
            pat_mean -= np.nanmin(pat_mean)
        pat_med = np.nanmedian(pat_mean)
        
    mean_thr = pat_med * mean_times
    diff_thr = pat_med * diff_times

    pat_mean[np.isnan(pat_mean)] = 0
    is_pat = pat_mean > mean_thr
    if (is_pat.any() == False) and (plot_norfi == True):
        if pdf is not None:
            plt.switch_backend('agg')
        fig,ax = plt.subplots(figsize = (15,3))
        ax.plot(t,pat_mean)
        ax.axhline(mean_thr,c = 'k',linestyle = '--',label='mean_thr')
        ax.grid()
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        if frange is not None:
            ax.set_title(f'freq in {frange} MHz')
        if xylim is not None:
            ax.set_xlim(xylim[0],xylim[1])
            ax.set_ylim(xylim[2],xylim[3])
        if pdf is not None:
            pdf.savefig();plt.close()
        return  is_timerfi

#     try:
    start,end = get_startend(is_pat,rfi_width_lim = rfi_width_lim,ext_sec=0)
#     except ValueError as V:
#         #import traceback
#         #traceback.print_exc()
#         #print(f"{V}.Not found.")
#         return  is_timerfi

    pat_diff = np.abs(np.diff(pat_mean, prepend=0, append=0))
    s,e = get_startend(is_timerfi, exclude = False)
    if len(start) > 0:
        print(f"rfi starts at tn = {start}, ends in tn = {end}")
        for s,e in zip(start,end):
            cond = (pat_diff[s] > diff_thr) & (pat_diff[e] > diff_thr)
            if cond:
                is_timerfi[s:e] = True

        if ext_add > 0 and is_timerfi.any():
            from ..utils.misc import extend_Trues
            is_timerfi = extend_Trues(is_timerfi,axis = 0,ext_frac = 0,ext_add = ext_add)
            s,e = get_startend(is_timerfi, exclude = False)
            print(f"After extension, rfi starts at tn = {s}, ends in tn = {e}")
#     else:
#         print("No True meets width condition.")
    if axis == 'time':
        if is_timerfi[1] == True: is_timerfi[0] = True
        if is_timerfi[-2] == True: is_timerfi[-1] = True

    if plot:
        if pdf is not None:
            plt.switch_backend('agg')
        print(f"Median value is {pat_med}. mean_thr = {mean_thr}. diff_thr = {diff_thr}")
        fig,ax = plt.subplots(figsize = (15,3))
        ax.plot(t,pat_mean)
        ax.axhline(mean_thr,c = 'k',linestyle = '--',label='mean_thr')
        ax.plot(t,is_timerfi*np.max(pat_mean),label='is_timeRFI')
        offset = np.max(pat_mean)
        ax.axhline( - offset,c = 'k',linestyle = '-')
        ax.plot(t,pat_diff[:-1] - offset,label='abs(diff)')
        ax.axhline(diff_thr - offset,c = 'k',linestyle = '--',label='diff_thr')

        ax.grid()
        ax.plot(t[s],pat_mean[s],'r.',label = 'start')
        ax.plot(t[e],pat_mean[e],'g.',label = 'end')
        ax.legend()
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        if frange is not None:
            ax.set_title(f'freq in {frange} MHz')
        if xylim is not None:
            ax.set_xlim(xylim[0],xylim[1])
            ax.set_ylim(xylim[2],xylim[3])
        if pdf is not None:
            pdf.savefig();plt.close()
    return  is_timerfi

def mask_sf(data,freq,frange = None,rms_frange = None,ext_times=1,mask_rms_times=3, is_timerfi = None, **kwargs):
    """
    Other parameters are the same as previous.
    """
    ret = np.zeros_like(data,dtype = 'bool')

    f_use = (freq>frange[0])&(freq<frange[1])
    if is_timerfi is None:
        is_timerfi = find_t(data,freq,frange = frange,**kwargs)

    if is_timerfi.any() == False:
        return ret
    log.info(f"Looking for short-freq time RFI in {frange} ...")

    start,end = get_startend(is_timerfi,ext_sec = 1)
    from ..utils.misc import smooth1d
    for s,e in zip(start,end):
        mspec = np.nanmean(data[s:e,:],axis = 0)
        sm = smooth1d(mspec[f_use],method='gaussian',sigma=10)
        MAX = max(sm)
        fmax = freq[f_use][np.nanargmax(sm)]
        f50 = freq[f_use][sm > MAX / 2]
        if len(f50) > 1:
            w50 = f50[-1] - f50[0]
            rms_thresh = rms(mspec,freq,rms_frange) * mask_rms_times

            mask_frange = find_edge_2sides(mspec,freq,peak_position=fmax-w50/2,step=w50,
                     rms_thresh=rms_thresh,Print=False,ext_times=ext_times,small_rfi_times=0)
            print(f"tn = {[s,e]} mask frange:",mask_frange)

            mask_use =  (freq>mask_frange[0])&(freq<mask_frange[1])
            ret[s:e,mask_use] = True
        else:
            ret[s:e,:] = True

    ext_add = kwargs['ext_add']
    if ext_add is not None:
        from ..utils.misc import extend_Trues
        ret = extend_Trues(ret,axis = -1,ext_frac = 0,ext_add = ext_add)

    print("Found :D")
    return ret

def mask_lf(data,freq, frange = None, rms_frange = None, mask_rms_times = -1,**kwargs):
    """
    Other parameters are the same as previous.
    """
    ret = np.zeros_like(data,dtype = 'bool')

    is_timerfi = find_t(data,freq,frange =frange, **kwargs)
    if is_timerfi.any() == False:
        print("No long-freq time rfi is found.")
        return ret
    log.info(f"Looking for long-freq time RFI ...")

    if mask_rms_times == -1:
        ret[is_timerfi,:] = True
        print("Mask the whole spec with rfi :(")
    elif mask_rms_times == 0:
        mask_use = (freq>frange[0])&(freq<frange[1])
        ret[is_timerfi,:] = True
        ret[:,~mask_use] = False
        print(f"Mask spec with rfi in input freq range {frange}")
    elif mask_rms_times > 0:
        # compute rms
        data2 = deepcopy(data)
        data[is_timerfi] = np.nan
        tmin = np.nanargmin(np.nanmean(data2, axis = 1))
        spec = data[tmin]
        
        rms_thresh = rms(spec,freq,rms_frange) * mask_rms_times
        mask_use = data[is_timerfi,:] > rms_thresh
        ret[is_timerfi,:] = mask_use
        print(f"Mask spec with rfi above rms thr {rms_thresh}")
    else:
        raise ValueError("mask_rms_times should be -1, 0, or >0 !")

    ext_add = kwargs['ext_add']
    if ext_add is not None:
        from ..utils.misc import extend_Trues
        ret = extend_Trues(ret,axis = -1,ext_frac = 0,ext_add = ext_add)

    print("Found :D")
    return ret

# Cell
def mask_time_rfi(data,freq,rtype = 'short-freq',frange = None, file=None,
                  frange_step = None, plot_norfi = False, **kwargs):
    """
    rtype: 'short-freq','long-freq'
    frange: freq range
    file: freq range exists short-freq time rfi npy filename
    frange_step: if frange is None and file is None, cycle in whole freq band
    Other parameters are the same as previous.
    """

    ret = np.zeros_like(data,dtype = 'bool')

    if rtype == 'short-freq':
        if file is not None or frange_step is not None:
            if file is not None:
                if not os.path.exists(file):
                    raise ValueError(f"can not find the input file path: {file}")
                franges = np.load(file)
                if franges.shape[1] != 2:
                    raise ValueError("time RFI freq shape must like (n,2)")
            else:
                frq1 = np.arange(freq[0],freq[-1],frange_step//2)[:-1]
                frq2 = np.hstack((np.arange(freq[0]+frange_step,freq[-1],frange_step//2),freq[-1]))
                franges = np.vstack((frq1,frq2)).T

            for nf in tqdm(range(franges.shape[0])):
                ret |= mask_sf(data,freq,frange = franges[nf],**kwargs)

        elif len(frange) == 2 and (frange[1] - frange[0]) > 0:
            ret = mask_sf(data,freq,frange,**kwargs)
        else:
            raise ValueError("frange should like [fmin,fmax] or a string (RFI npy filepath) or a certain frange_step.")

        if ret.any() == False:
            print("No short-freq time rfi is found.")

    elif rtype == 'long-freq':
        if len(frange) == 2:
            ret = mask_lf(data,freq, frange,  **kwargs)
        else:
            raise ValueError("frange should like [fmin,fmax].")

    print("Finish")
    return ret

def mask_nr(data,freq, rms_frange = None, mask_rms_times = 0, **kwargs):
    """
    Other parameters are the same as previous.
    """
    ret = np.zeros_like(data,dtype = 'bool')

    log.info(f"Looking for long-time narrowband freq RFI ...")
    is_freqrfi = find_t(data,freq,axis = 'freq', **kwargs)
    if is_freqrfi.any() == False:
        print("No long-time freq rfi is found.")
        return ret

    if mask_rms_times == 0:
        ret[:,is_freqrfi] = True
        print("Mask the whole chanels with rfi :P")
    elif mask_rms_times > 0:
        # compute rms
        tmin = np.nanargmin(np.nanmean(data, axis = 1))
        spec = data[tmin]
        rms_thresh = rms(spec,freq,rms_frange) * mask_rms_times
        mask_use = data[:,is_freqrfi] > rms_thresh
        ret[:,is_freqrfi] = mask_use
        print(f"Mask channels with rfi above rms thr {rms_thresh}")

    ext_add = kwargs['ext_add']
    if ext_add is not None:
        from ..utils.misc import extend_Trues
        ret = extend_Trues(ret,axis = 0,ext_frac = 0,ext_add = ext_add)

    print("Found :D")
    return ret

# Cell
def mask_freq_rfi(data,freq,rtype = 'long-time',**kwargs):
    if rtype == 'long-time':
        ret = mask_nr(data,freq, **kwargs)

    print("Finish")
    return ret

def mask_thr_rfi(T2p, freq, s_sigma = 5, rms_frange = None, times=6, ext_add=0, ext_frac=0):
    """
    Mask RFI based on threshold detection after Gaussian smoothing.
    
    Parameters
    ----------
    T2p : array
        Array of shape (time, freq, 2); T or flux with two polarizations sorted by time.
    freq : array
        Frequency array corresponding to the second axis of T2p.
    s_sigma : float, optional
        Sigma for Gaussian smoothing along frequency axis. Default is 5.
    rms_frange : list or None, optional
        Frequency range [fmin, fmax] for RMS calculation. If None, use all frequencies.
    times : float, optional
        Threshold multiplier. RFI is identified where T > RMS * times. Default is 6.
    ext_add : int, optional
        Number of additional channels to extend on each side of detected RFI. Default is 0.
    ext_frac : float, optional
        Fractional extension of RFI regions. Default is 0.
    
    Returns
    -------
    is_rfi : ndarray
        Boolean array of shape (time, freq) indicating RFI positions.
    
    """
    from scipy import ndimage
    from ..utils.misc import extend_Trues
    
    # import time  # Add this at the top of the file if not already imported
    # start_time = time.time()  # Start timing

    T = np.nanmean(T2p, axis = 2) # shape (time, freq)
    
    from scipy import ndimage
    T = ndimage.gaussian_filter1d(T, s_sigma, axis=1) # smooth along freq axis

    RMS = rms(T,freq,rms_frange)[:, None]

    # find rfi
    is_rfi = (T > RMS * times) 
    
    if ext_add > 0 or ext_frac > 0:
        is_rfi = extend_Trues(is_rfi, axis=1, ext_add=ext_add, ext_frac=ext_frac)

    # end_time = time.time()  # End timing
    # print(f"mask_thr_rfi execution time: {end_time - start_time:.2f} seconds")

    return is_rfi





__all__ = []


import h5py
import warnings


from .core.cal import CalOnOff
from .core.cal import FastRawSpec
from .core.radec import get_radec

from .core.flux import FluxCali

from .core.corr_vel import frame_correct_freq
from .core.corr_vel import freq2vel, vel2freq

from .utils.io import MjdChanPolar_to_PolarMjdChan
from .utils.io import PolarMjdChan_to_MjdChanPolar
from .utils.io import get_nB
from .utils.io import replace_nB
from .utils.io import load_hdf5_to_dict

try:
    from .interaction import bld_i as interact_bld
    from .interaction import sw_i as interact_sw
except ImportError as err:
    print(err)
    warnings.warn(f"{err}. Failed to import interactive modules")

from .utils.io import HFDataT

from .utils import obs_log


class HFSpec:

    def __init__(self, s, *,
                 kT=('Ta', 'flux', 'Power', 'DATA'),
                 DATA_cand=('Ta', 'flux', 'Power')):
        """
        s: str
        kT: field needed to be transposed by PolarMjdChan_to_MjdChanPolar
        DATA_cand: key candidates as 'DATA'
        """
        if isinstance(s, str):
            self.fin = h5py.File(s, 'r')
        else:
            raise(ValueError('input'))
        self._group = self.fin['S']
        self._keys_trans_ = kT
        self._DATA_cand = DATA_cand

        self._gen_DATA()
        self._set_attrs()


    def _gen_DATA(self,):
        self._DATA = None
        keys_ = self._group.keys()
        if 'DATA' in keys_:
            self._DATA = 'DATA'
        else:
            for key in self._DATA_cand:
                if key in keys_:
                    self._DATA = key
                    break
        # if self._DATA is None:
        #     raise(KeyError('can not find spec data'))

    def keys(self,):
        keys = list(self._group.keys())
        if self._DATA is not None and 'DATA' not in keys:
            keys.append('DATA')
        return tuple(keys)

    @property
    def Header(self,):
        Header = dict(self.fin['Header'].attrs.items())
        return Header

    def __getitem__(self, arg):
        if arg == 'DATA':
            arg = self._DATA
        r = self._group.__getitem__(arg)
        if arg in self._keys_trans_:
            r = HFDataT(r)
        if r.ndim == 1:
            r = r[:]
        return r

    def __repr__(self,):
        return f"keys:\n{repr(self.keys())}"

    def _set_attrs(self,):
        keys_forbid = ['keys', 'fin']
        for key in self.keys():
            if key in keys_forbid:
                warnings.warn(f'can not set {key} as attrs')
            else:
                setattr(self, key, self.__getitem__(key))





__all__ = ['get_ratio', 'ATemp_correction', 'download_flux_gain_Liu2024', 'get_flux_gain_Liu2024', 'Calibrator',
           'FluxCali']



import os
import re
from glob import glob
import warnings

import h5py
import numpy as np
import scipy.interpolate as interp
import json

from . import conf
from astropy.time import Time
from astropy import units

from .gain import Get_gain, gain_diff_from_ZA, Get_ZA
from ..utils.downloader import get_file


def get_ratio(nB, freq=None):
    """get gain ratio to beam 1 in arXiv:2002.01786"""
    
    with open(os.path.dirname(__file__) + '/data/beam_ratios.json', 'r') as f:
        ratios_para = json.load(f)

    # Assumes JSON structure: "M01": {"1050": val, "1050_err": err, ...} (ordered)
    # Original: freq_key = np.array(ratios_para.index[::2], dtype=float)
    # ratios = ratios_para[f'M{nB:02d}'][::2].values
    
    beam_data = ratios_para[f'M{nB:02d}']
    # Keys/Values are ordered in Python 3.7+
    keys = list(beam_data.keys())
    values = list(beam_data.values())
    
    freq_key = np.array(keys[::2], dtype=float)
    ratios = np.array(values[::2])
    
    if freq is not None:
        ratios= interp.interp1d(freq_key,ratios, kind='quadratic', fill_value= "extrapolate")(freq)
        freq_key=freq
    return ratios, freq_key


## flux gain from Liu et al. 2024
def ATemp_correction(ATemp_pre, ATemp_target, freq):
    """
    ATemp: Ambient temperature (in degrees Celsius) of the pre-measured flux gains
    ATemp_target: Ambient temperature (in degrees Celsius) of the target observation.
    freq: frequency (in MHz) of the pre-measured flux gains
    """
    s_factor = 4.7e-5*freq - 1.4e-3 # equation from Liu et al. 2024
    K_Jy_diff = s_factor * (ATemp_target - ATemp_pre)
    return K_Jy_diff

def download_flux_gain_Liu2024():
    url = 'https://download.scidb.cn/download?fileId=5a612e328587f62c27a3435aac1af0fa'
    fpath = os.path.expanduser('~/.cache/hifast/data/flux_gain_Liu2024.hdf5')
    return get_file(url, fpath, max_retries=60, retry_delay=10)


def get_flux_gain_Liu2024(nB, ZA_target, ATemp_target=None, freq=None):
    """
    nB: scalar
        beam number
    ZA_target: array_like
        Zenith angle (in degrees) of the target observation
    ATemp_target: scalar
        Ambient temperature (in degrees Celsius) of the target observation
    freq: array_like
        frequency (in MHz) of the pre-measured flux gains

    """
    fpath = download_flux_gain_Liu2024()
    if fpath is None:
        raise(ValueError('can not obtain flux gain from Liu et al. 2024'))
    # load
    print(f'loading {fpath}')
    with h5py.File(fpath, 'r') as fs:
        freq_pre = fs['freq'][()]
        ATemp_pre = fs['Ambient_Temperature'][()]
        ZA_pre = fs['ZA'][()]
        K_Jy = fs[f'K_Jy{nB}'][()] # shape (1,n,1)

    if freq is not None:
        K_Jy = interp.interp1d(freq_pre, K_Jy, kind='linear', fill_value= "extrapolate", axis=1)(freq)
        freq_sample = freq
    else:
        freq_sample = freq_pre

    if ATemp_target is not None:
        print(f'Ambient temperature correction from {ATemp_pre} degree C to the target {ATemp_target} degree C')
        K_Jy_diff = ATemp_correction(ATemp_pre, ATemp_target, freq_sample)
        K_Jy += K_Jy_diff[None, :, None]
    else:
        warnings.warn(f'no ambient temperature correction')
    # Zenith angle correction
    if ZA_target is not None:
        K_Jy_diff = gain_diff_from_ZA(ZA_pre, ZA_target, nB, freq_sample)
        K_Jy = K_Jy + K_Jy_diff[..., None]
        print(f'Zenith angle correction from {ZA_pre} degree to the target {ZA_target} degree')
    else:
        warnings.warn(f'no Zenith angle correction')

    return K_Jy, freq_sample


class Calibrator:

    def __init__(self, mjd, nB, cbr_store, cbr_name='*', use_counts=True, only_use_19beams=False):
        """
        mjd: used to find nearest cbr
        nB: beam number
        cbr_store: str
           where quasar calibrator stored in, a file path or a directory including several files
           If None, use the gain depended on Zenith angle (arxiv:2002.01786) and need input ra, dec and mjd.
        cbr_name: str, e.g. 3C48
           If not None and cbr_store is a directoy, searching the specified calibrator.
        use_counts: bool
           If True, return gain as counts/Jy, If False, K/Jy
        """
        self.cbr_store = cbr_store
        self.cbr_name = cbr_name
        self.mjd = mjd
        self.nB = nB
        self.use_counts = use_counts
        self.only_use_19beams = only_use_19beams
        self.parser_cbr()

    @staticmethod
    def get_date_nearest(fpath_list, mjd):
        dates_have = [os.path.basename(fpath).rsplit('-FluxGain', 1)[0].rsplit('-', 1)[-1] for fpath in fpath_list]
        dates_have = [f'{s[:4]}-{s[4:6]}-{s[6:8]} {s[8:10]}:{s[10:12]}:00.000' for s in dates_have]
        # date of file name is UTC
        mjds_h = Time(dates_have, format='iso', scale='utc').mjd
## use mjd
#         mjds_h = np.empty(len(fpath_list), dtype='float64')
#         for i, fpath in enumerate(fpath_list):
#             with h5py.File(fpath) as fs:
#                 mjds_h[i] = fs['mjd'][0]
        ind = np.argmin(abs(mjds_h - mjd))
        return fpath_list[ind]

    def _calc_ratio(self, fpath):
        """calc gain ratio to beam 1"""
        with h5py.File(fpath, 'r') as fs:
            freq = fs['freq'][()]
            K_Jy = fs[f'K_Jy{self.nB}'][0] if not self.use_counts else fs[f'count_Jy{self.nB}'][0]
            K_Jy_nB1 = fs[f'K_Jy1'][0] if not self.use_counts else fs[f'count_Jy1'][0]
#             print(K_Jy,K_Jy_nB1)
            ratio = K_Jy/K_Jy_nB1
            ZA_nB = self.get_ZA_cbr(fs, self.nB)
        return freq, ratio, ZA_nB

    def get_ratio(self):
        # cbr_store is directory
        fpaths_nB_all = glob(os.path.join(self.cbr_store, f"**/{self.cbr_name}-20*-FluxGain-All.hdf5"), recursive=True)
        if len(fpaths_nB_all) == 0:
            raise(ValueError('can not find 19-beams calibration file to calculate relative gain'))
        else:
            fpath = self.get_date_nearest(fpaths_nB_all, self.mjd)
            print(f"using {fpath} to calculate relative gain to M01")
            freq, ratio, ZA_nB = self._calc_ratio(fpath)
        return freq, ratio, ZA_nB

    def parser_cbr(self,):
        """assign or serach fpath from self.cbr_store"""
        if not os.path.exists(self.cbr_store):
            raise(ValueError(f'can not find the calibrator: {self.cbr_store}'))
        if os.path.isfile(self.cbr_store):
            self.cbr_fpath = self.cbr_store
        else:
            if self.only_use_19beams:
                fpaths_cand = glob(os.path.join(self.cbr_store, f"**/{self.cbr_name}-20*-FluxGain-All.hdf5"), recursive=True)
            else:
                fpaths_cand = glob(os.path.join(self.cbr_store, f"**/{self.cbr_name}-20*-FluxGain*.hdf5"), recursive=True)
            if len(fpaths_cand) == 0:
                raise(ValueError(f"can not find calibrator file"))
            self.cbr_fpath = self.get_date_nearest(fpaths_cand, self.mjd)

    def get_ZA_cbr(self, fs, nB):
        if f'ZA_cbr{nB}' in fs.keys():
            ZA = fs[f'ZD_cbr{nB}'][()]
        else:
            ZAs = fs[f'ZD{nB}'][()]
            ZA = np.mean(ZAs)
        # need deg
        ZA = np.rad2deg(ZA)
        return ZA

    def __call__(self, freq_new=None, ZAs_obs=None, tcal_spec=None):
        """
        freq_new: if not None, will interplate
        ZAs_obs: ZA of obs spec. If not None, will fix the gain diff from ZA diff
        """
        print(f'loading gain from {self.cbr_fpath}')
        with h5py.File(self.cbr_fpath) as fs:
            freq = fs['freq'][()]
            key_ = f'K_Jy{self.nB}' if not self.use_counts else f'count_Jy{self.nB}'
            if freq_new is None:
                freq_new = freq
            if key_ in fs.keys():
                gain = fs[key_][()] # K/Jy, shape:(1, chan, 2)
                gain = interp.CubicSpline(freq, gain, axis=1, extrapolate=True)(freq_new)
            else:
                # use M01 or relative gain  ratio
                gain_nB1 = fs[f'K_Jy1'][()] if not self.use_counts else fs[f'count_Jy1'][()]
                gain_nB1 = interp.CubicSpline(freq, gain_nB1, axis=1, extrapolate=True)(freq_new)

                freq_ratio, ratio, ZA_cbr_nB_ratio = self.get_ratio()
                ratio = interp.CubicSpline(freq_ratio, ratio, extrapolate=True)(freq_new)
                gain = gain_nB1 * ratio[None,:]

            if self.use_counts:
                # multiply count with spectra's tcal
                gain *= tcal_spec # to K/Jy

            if ZAs_obs is not None:
                # fix ZA diff
                ZA_cbr = self.get_ZA_cbr(fs, self.nB) if key_ in fs.keys() else ZA_cbr_nB_ratio
                diff = gain_diff_from_ZA(ZA_cbr, ZAs_obs, self.nB, freq_new)
                gain = gain + diff[..., None]
                print(f'fixing ZA diff from ZA_cbr:{ZA_cbr} to ZA_obs:{ZAs_obs}')
        return freq_new, gain



class FluxCali:

    def __init__(self, nB, freq, cbr_store=None, cbr_name='*', tcal_spec=None,  only_use_19beams=False,
                           pre_measured='Jiang2020', ATemp=None,
                           mjd=None, ra=None, dec=None, ZAs=None, fix_diff_ZA=False):
        """

        flux calibration using Calibator or Table Gain
        -----------------------
        T: array_like
           Temperature of the spectra. Shape is (m,n) or (m,n,2) i.e. (Mjd, channel) or (Mjd, channel, Polarization)
        nB: int
           Beam numbe
        freq: array_like, shape (n,)
        cbr_store: str
           where quasar calibrator stored in, a file path or a directory including several files
           If None, use the gain depended on Zenith angle (arxiv:2002.01786) and need input ra, dec and mjd.
        cbr_name: str, e.g. 3C48
           If not None and cbr_store is a directoy, searching the specified calibrator.
        tcal_spec: shape: (1,n,2); T.shape need be (m,n,2); if spectra and Calibator temperature used different Tcal, input this to fix it.
        ra, dec, mjd: None or array_like, shape (m,)

        pre_measured: str
           'Jiang2020' or 'Liu2024'
        Atemp: array_like
           Ambient temperature (in degrees Celsius) of the target observation

        """

        if ra is not None and dec is not None:
            if np.isscalar(mjd):
                mjd = np.array([mjd])
            if np.isscalar(ra):
                ra = np.array([ra])
            if np.isscalar(dec):
                dec = np.array([dec])
            assert ra.shape[0] == dec.shape[0] == mjd.shape[0]

        self.nB = nB
        self.freq = freq
        self.cbr_store = cbr_store
        self.cbr_name = cbr_name
        self.tcal_spec = tcal_spec
        self.only_use_19beams = only_use_19beams
        self.mjd = mjd

        self.ra = ra
        self.dec = dec
        self.ZAs = ZAs
        self.fix_diff_ZA = fix_diff_ZA

        self.ATemp = ATemp

        if cbr_store is None or cbr_store=='none':
            if pre_measured == 'Jiang2020':
                print('Using pre-measured flux gain from Jiang2020')
                self.K_Jy = self.get_gain_tabled()
            elif pre_measured == 'Liu2024':
                print('Using pre-measured flux gain from Liu2024')
                self.K_Jy = self.get_gain_Liu2024()
            else:
                raise ValueError(f'pre_measured {pre_measured} not supported')
        else:
            print(f'Using Calibrator from {cbr_store}')
            self.K_Jy = self.get_gain_cbr()

    def get_gain_cbr(self,):

        use_counts = True if self.tcal_spec is not None else False

        CBR = Calibrator(np.mean(self.mjd), self.nB, self.cbr_store,
                         use_counts=use_counts,
                         only_use_19beams=self.only_use_19beams)
        if self.fix_diff_ZA:
            if self.ZAs is not None:
                ZAs = self.ZAs
            else:
                ZAs = Get_ZA(self.ra, self.dec, self.mjd)
        else:
            ZAs = None
        _, gain = CBR(self.freq, ZAs, self.tcal_spec)
        self.cbr_fpath = CBR.cbr_fpath
        K_Jy = gain
        return K_Jy


    def get_gain_tabled(self,):
        """from Jiang2020"""
        K_Jy = Get_gain(self.ra, self.dec, self.mjd, self.nB, self.freq)[0]  # K/Jy

        return K_Jy[:, :, None]

    def get_gain_Liu2024(self,):
        """from Liu2024"""
        if self.ZAs is not None:
            ZA_target = self.ZAs
        else:
            ZA_target = Get_ZA(self.ra, self.dec, self.mjd)
        K_Jy, _ = get_flux_gain_Liu2024(self.nB, ZA_target=ZA_target, ATemp_target=self.ATemp, freq=self.freq)
        return K_Jy

    def __call__(self, T):
        """
        T: array; shape: (Mjd,Chan,Polar) or (Mjd,Chan)
        """
        K_Jy = self.K_Jy
        # cali
        if T.ndim == 2:
            return T/np.mean(K_Jy, axis=2)
        elif T.ndim == 3:
            if T.shape[-1] == 2:
                return T/K_Jy
            elif T.shape[-1] == 1:
                return T/np.mean(K_Jy, axis=2, keepdims=True)
            else:
                raise(ValueError('shape of T'))
        else:
            raise(ValueError('shape of T'))



import configargparse
import argparse


class write_config(object):
    def __init__(self, namespace):
        self.namespace = namespace

    def write_action(self, a):

        is_ = False
        for item in a.option_strings:
            is_ = is_ or item.startswith('--')
            is_ = is_ and item != '--help'
        is_ = is_ and not getattr(a, "not_in_write_out_config_file", False)
        is_ = is_ and not getattr(a, "is_config_file_arg", False)
        #is_ = is_ and not getattr(a, "is_write_out_config_file_arg", False)

        if is_:
            msg = '#help: '
            msg += f"{', '.join(a.option_strings)}; "
            msg += f"default: {a.default}; " if a.default is not None else 'default: ;'
            if a.choices is not None:
                msg += f"choices: {', '.join(map(str,a.choices))}; "
            msg += '\n# '.join(a.help.split('\n'))
            value = getattr(self.namespace, a.dest, '')
            if isinstance(value, list):
                value = "["+", ".join(map(str, value))+"]"
            if value is None:
                msg += f"\n#{a.dest} = "
            elif isinstance(a.const, bool):
                msg += f"\n{a.dest}" if a.const == value else f"\n#{a.dest}"
            elif isinstance(value, bool) and a.default.lower() in ['yes', 'no']:
                msg += f"\n{a.dest} = yes" if value else f"\n{a.dest} = no"
            else:
                msg += f"\n{a.dest} = {value}"
            return msg+'\n'
        else:
            return ''

    def write_group(self, g):
        msg = ''
        for a in g._group_actions:
            msg += self.write_action(a)
        if msg != '':
            title = '\n## '.join(g.title.split('\n'))
            msg = f"## {title}\n" + msg + '\n'
        return msg

    def write_groups(self, gs):
        msg = ''
        for g in gs:
            msg += self.write_group(g)
        return msg

def add_argument(self, *args, **kwargs):
    """
    Arguments added:
        not_in_write_out_config_file: excluded in 'is_write_out_config_file_arg' file
        not_add_hyphen_option: bool
    """
    not_in_write_out_config_file = kwargs.pop("not_in_write_out_config_file", None)
    not_add_hyphen_option = kwargs.pop("not_add_hyphen_option", False)
    
    if not not_add_hyphen_option:
        args = list(args)
        for op in args:
            if '_' in op and op.startswith('-'):
                op_hy = op.replace('_', '-')
                if op_hy not in args:
                    args += [op_hy, ]
    
    action = self.configargparse_original_add_argument(*args, **kwargs)
    action.not_in_write_out_config_file = not_in_write_out_config_file
    
    
    
    return action


add_argument.__doc__ += configargparse.ArgParser.add_argument.__doc__


class ArgumentParser(configargparse.ArgParser):
    def write_config_file(self, parsed_namespace, output_file_paths, exit_after=False):
        if output_file_paths:
            ww = write_config(parsed_namespace)
            for output_file_path in output_file_paths:
                with open(output_file_path,'w') as f:
                    f.write(ww.write_groups(self._action_groups))
            print("Wrote config file to " + ", ".join(output_file_paths))
            if exit_after:
                self.exit(0)
                
        # del argument with not_in_write_out_config_file in parsed_namespace
        #[delattr(parsed_namespace, a.dest) for a in self._actions if getattr(a, "not_in_write_out_config_file", False)]
        #super().write_config_file(parsed_namespace, output_file_paths, exit_after)
    #write_config_file.__doc__ = configargparse.ArgParser.write_config_file.__doc__

argparse._ActionsContainer.configargparse_original_add_argument = configargparse.ArgParser.add_argument
argparse._ActionsContainer.add_argument = add_argument
argparse._ActionsContainer.add_arg = add_argument
argparse._ActionsContainer.add = add_argument

ArgParser = ArgumentParser
Parser = ArgumentParser





__all__ = ['PolarMjdChan_to_MjdChanPolar', 'MjdChanPolar_to_PolarMjdChan', 'Test', 'phrase_ylim', 'main']


import numpy as np
from scipy import ndimage
import matplotlib.pyplot as plt
import h5py

import re
import os
from glob import glob
import json

from ..core.baseline import get_baseline, sub_baseline
from ..utils import io
from .widgets import *

import warnings
warnings.filterwarnings("ignore", r'overflow encountered in exp')

import matplotlib as mpl

PolarMjdChan_to_MjdChanPolar = lambda x:io.PolarMjdChan_to_MjdChanPolar(x, check_polar=False)
MjdChanPolar_to_PolarMjdChan = lambda x:io.MjdChanPolar_to_PolarMjdChan(x, check_polar=False)


class Test(object):
    def __init__(self, T2p, freq, frange=None, is_excluded=None, trans=False, save=False):
        """
        T2p: shape (Polar,Mjd,Chan)

        """
        self.trans = trans
        if T2p.ndim != 3:
            raise(ValueError('T2p should be 3 dim'))
        self.T2p = T2p

        self.is_excluded = is_excluded

        freq = freq[:]
        if self.trans:
            freq = np.arange(T2p.shape[1])
        if frange is not None:
            self.is_use = (freq > frange[0]) & (freq < frange[1])
            self.freq = freq[self.is_use]
        else:
            self.is_use = None
            self.freq = freq
        self.select()

        if save:
            # increase memory usage
            # (Mjd,Chan,Polar)
            self.bld_out = np.full((self.T2p.shape[1], len(self.freq), self.T2p.shape[0]), np.nan)
        else:
            self.bld_out = None
        self.save = save

    def select(self, start=0, length=20, polar=0):
        """
        select one polar of spectra from `start` to `start + length`
        """
        if self.trans:
            self.T2p_t = self.T2p[polar:polar+1,:, start:start+length]
            self.T2p_t = PolarMjdChan_to_MjdChanPolar(self.T2p_t)
            self.T2p_t = self.T2p_t.transpose((1, 0, 2))
        else:
            self.T2p_t = PolarMjdChan_to_MjdChanPolar(self.T2p[polar:polar+1, start:start+length])
        # also change self.is_excluded
        if self.is_excluded is not None:
            if self.is_excluded.shape == self.T2p.shape:
                if self.trans:
                    self.is_excluded_t = self.is_excluded[polar:polar+1, :, start:start+length]
                    self.is_excluded_t = PolarMjdChan_to_MjdChanPolar(self.is_excluded_t).transpose((1, 0, 2))
                else:
                    self.is_excluded_t = PolarMjdChan_to_MjdChanPolar(self.is_excluded[polar:polar+1, start:start+length])
            elif self.is_excluded.ndim == 2 and self.is_excluded.shape == T2p.shape[1:]:
                if self.trans:
                    self.is_excluded_t = self.is_excluded[:, start:start+length][None,:]
                    self.is_excluded_t = PolarMjdChan_to_MjdChanPolar(self.is_excluded_t).transpose((1, 0, 2))
                else:
                    self.is_excluded_t = self.is_excluded[start:start+length][..., None]
            else:
                raise(ValueError('shape of ``is_excluded``'))
        else:
            self.is_excluded_t = None

        if self.is_use is not None:
            self.T2p_t = self.T2p_t[:, self.is_use]
            if self.is_excluded_t is not None:
                self.is_excluded_t = self.is_excluded_t[:, self.is_use]

    def sub(self, x, start=0, length=20, polar=0, frange_excluded=None,
            exclude_add='none', verbose=False, **kwargs):

        #store
        self.start = start
        self.length = length
        self.polar = polar

        self.select(start, length, polar)

        if frange_excluded is not None and frange_excluded[0] != frange_excluded[1]:
            is_ = (x > frange_excluded[0]) & (x < frange_excluded[1])
            # full
            is_ = np.full(self.T2p_t.shape, is_[None,...,None])
            if self.is_excluded_t is not None:
                self.is_excluded_t &= is_
            else:
                self.is_excluded_t = is_

        self.frange_excluded = frange_excluded
        self.sub_baseline_para = kwargs
        self.sub_baseline_para['is_excluded'] = self.is_excluded_t
        self.sub_baseline_para['exclude_add'] = exclude_add
        self.sub_baseline_para['knots'] = json.loads(self.sub_baseline_para['knots'])

        self.bld = sub_baseline(self.freq, self.T2p_t, verbose=verbose, **self.sub_baseline_para)
        if self.bld_out is not None:
            self.bld_out[start:start+length, :, polar:polar+1] = self.bld

        return np.full(x.shape, np.nan)

    def get_ori(self, x, i, **kwargs):
        res = self.T2p_t[i, :, 0]
        return np.vstack([res, ndimage.gaussian_filter1d(res, 3)]).T

    def get_bld(self, x, i, **kwargs):
        res = self.bld[i, :, 0]
        return np.vstack([res, ndimage.gaussian_filter1d(res, 3)]).T

    def get_bl(self, x, i,  **kwargs):
        return self.T2p_t[i, :, 0] - self.bld[i, :, 0]

    def get_bld_mean(self, x, start_stop,  **kwargs):
        start, stop = start_stop
        res = np.mean(self.bld[int(start):int(stop)+1, :, 0], axis=0)
        return np.vstack([res, ndimage.gaussian_filter1d(res, 3)]).T


def phrase_ylim(ylim, vals):
    is_ = np.isfinite(vals)
    ylim = list(map(lambda x: float(x) if 'per' != str(x)[:3] else
                    np.nanpercentile(vals[is_], float(str(x)[3:]), interpolation='nearest'), ylim))
    return ylim



def main():
    global ylim

    tes = Test(T2p, freq, frange, is_excluded, trans, save)

    rew_type = ['asym1', 'asym2', 'asym3', 'sym1',]
    method = ['PLS-'+r for r in rew_type]
    method += ['poly-'+r for r in rew_type]
    method += ['Gauss-'+r for r in rew_type]
    method += ['knspline-'+r for r in rew_type]
    method += ['knpoly-'+r for r in rew_type]
    method += ['spline-'+r for r in rew_type]
    method += ['masPLS-'+r for r in rew_type]
    method += ['asPLS',]

    sliders = {}
    if trans:
        start_e = tes.T2p.shape[2]-length
    else:
        start_e = tes.T2p.shape[1]-length
    if start_init is None:
        start_i = start_e//2
    else:
        start_i = min(start_init, start_e)
    sliders.update(_BoundedIntText(
                       start=(start_i, 0, start_e, 1), polar=(0, 0, tes.T2p.shape[0], 1)))
    sliders.update(_IntSlider(njoin=(0, 1, length, 1),
                              average_every_freq=(0, 1, 40, 1)))
    sliders.update(_Dropdown(method=method,
                             s_method_freq=('none', 'gaussian', 'boxcar',),
                             s_method_t=('none', 'gaussian', 'boxcar'),
                             exclude_add=('none', 'auto1', 'auto2'),
                             ))
    sliders.update(_IntSlider(s_sigma_freq=(3, 1, 20, 1), s_sigma_t=(3, 1, 20, 1)))
    sliders.update(_FloatLog10Slider(lam=(1e8, 0, 13, 0.2), readout_format='.2e'))
    sliders.update(_IntSlider(deg=(2, 1, 10, 1),
                              niter=(100, 1, 200, 1)))
    sliders.update(_FloatSlider(offset=(2, 0.1, 4, 0.1)))
    sliders.update(_FloatSlider(ratio=(0.01, 0.001, 0.015,0.001),
                                readout_format='.3f'))

    sliders['knots'] = widgets.Text(
                    value=json.dumps(tes.freq[[1,-2]].tolist()),
                    description='knots',
                    layout=widgets.Layout(width=f"{mpl.rcParams['figure.dpi']*figsize[0]*1.2}px",),
                    disabled=False)


    bak = w_conf.pop('layout')
    #w_conf['style'] = {'description_width': 'initial'}
    w_conf['layout'] = widgets.Layout(width=f"{mpl.rcParams['figure.dpi']*figsize[0]*1.2}px",)
    sliders.update(_FloatRangeSlider(frange_excluded=([tes.freq[0], tes.freq[0]], tes.freq[0], tes.freq[-1], tes.freq[2]-tes.freq[0])))
    w_conf['layout'] = bak

    plt.ioff()
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=figsize)
    fig.canvas.header_visible = False

    # ylim xlim in plot
    if isinstance(ylim, str):
        if ylim == 'auto':
            ylim = phrase_ylim(['per0.1', 'per99.5'], PolarMjdChan_to_MjdChanPolar(T2p[:,:5]))
        elif ylim == 'full':
            ylim = 'fixed'
    else:
        ylim = phrase_ylim(ylim, tes.T2p_t)


    if isinstance(ylim, str):
        ylim2 = ylim
    else:
        ylim2 = [(ylim[0]-ylim[1])/2, (ylim[1]-ylim[0])/2]
    xlim = [tes.freq.min(), tes.freq.max()]

    widget_i = _IntSlider(i=(0, 0, length-1, 1))['i'] if length ==1 else range(length)


    # init
    controls = iplt.plot(tes.freq, tes.sub, nproc=nproc, length=length,
                         **sliders,
                         xlim=xlim, ylim=ylim,
                         color='r', ax=ax1, play_buttons=True, display_controls=False)
    controls2 = iplt.plot(tes.freq, tes.get_ori, i=widget_i,
                          controls=controls, xlim=xlim, ylim=ylim, ax=ax1, display_controls=False, play_buttons=True)
    controls3 = iplt.plot(tes.freq, tes.get_bl,
                          controls=controls2, xlim=xlim, ylim=ylim, ax=ax1)

    iplt.plot(tes.freq, tes.get_bld,
              controls=controls2, xlim=xlim, ylim=ylim2, ax=ax2)
    plt.axhline(y=0,)

    controls4 = iplt.plot(tes.freq, tes.get_bld_mean, start_stop=('r', 0, length, 11),
                          controls=controls, xlim=xlim, ylim=ylim2, ax=ax3)

    plt.axhline(y=0, )
    ax1.set_title('orange line: gaussian smoothed for better view. \n origial spectra', fontsize=8)
    ax2.set_title('spectra after baseline removed', fontsize=8)

    def return_v(*args, **kwargs): return tes.frange_excluded[0]
    [iplt.axvline(return_v, controls=controls, ax=ax, color='r') for ax in [ax1, ax2, ax3]]
    def return_v(*args, **kwargs): return tes.frange_excluded[1]
    [iplt.axvline(return_v, controls=controls, ax=ax, color='r') for ax in [ax1, ax2, ax3]]

    fig.tight_layout()

    w = controls.controls
    hbs = [
        widgets.HBox([widgets.Label(value=rf"$Select$ spectra from $start$(max:{controls.controls['start'].max}) to $start+{length}$ to fit:"),
                      w['start'], w['polar'],
                      widgets.Label('.'),
                      w['ratio']]),
        widgets.HBox(
            [widgets.Label(value=r"$One$ spectrum in top & middle:"),
             w['i'],
             widgets.Label(value=r"$Stacked$ spectra in bottom:"),
             w['start_stop']]),
        widgets.HBox([w['njoin'],
                      w['s_method_t'],
                      w['s_sigma_t'],
                      w['niter'],
                     ]),
        widgets.HBox([w['average_every_freq'],
                      w['s_method_freq'],
                      w['s_sigma_freq'],
                      w['offset']]),
        widgets.HBox([w['method'], w['lam'], w['deg'],
                      w['exclude_add']]),
        widgets.HBox([w['knots'], ]),
        widgets.HBox([w['frange_excluded'], ]),
    ]

    BOX = widgets.VBox(hbs)
    display(BOX)
    plt.show()
    display(BOX)

    return tes, w, BOX
